﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Globalization;
using System.IO;
using System.Linq;

namespace CSharpRecipes
{
    public static class Generics
    {
        #region "2.1 ジェネリックを使用するタイミングと場所を決定する"
        // 詳細については本書の該当レシピを参照してください
        #endregion

        #region "2.2 ジェネリック型を理解する"
        public static void TestGenericClassInstanceCounter()
        {
            // 標準的なクラス
            FixedSizeCollection A = new FixedSizeCollection(5);
            Console.WriteLine(A);
            FixedSizeCollection B = new FixedSizeCollection(5);
            Console.WriteLine(B);
            FixedSizeCollection C = new FixedSizeCollection(5);
            Console.WriteLine(C);

            // ジェネリッククラス
            FixedSizeCollection<bool> gA = new FixedSizeCollection<bool>(5);
            Console.WriteLine(gA);
            FixedSizeCollection<int> gB = new FixedSizeCollection<int>(5);
            Console.WriteLine(gB);
            FixedSizeCollection<string> gC = new FixedSizeCollection<string>(5);
            Console.WriteLine(gC);
            FixedSizeCollection<string> gD = new FixedSizeCollection<string>(5);
            Console.WriteLine(gD);

            bool b1 = true;
            bool b2 = false;
            bool bHolder = false;

            // 標準的クラスに要素を(objectとして)追加
            A.AddItem(b1);
            A.AddItem(b2);
            // ジェネリッククラスに要素を(boolとして)追加
            gA.AddItem(b1);
            gA.AddItem(b2);

            Console.WriteLine(A);
            Console.WriteLine(gA);

            // 返り値をキャストします。キャストしない場合はコンパイルエラーが発生します:
            // error CS0266: 型 'object' を 'bool' に暗黙的に変換できません...
            bHolder = (bool)A.GetItem(1);
            // キャストは必要ありません
            bHolder = gA.GetItem(1);

            int i1 = 1;
            int i2 = 2;
            int i3 = 3;
            int iHolder = 0;

            // 標準的クラスに要素を(objectとして)追加
            B.AddItem(i1);
            B.AddItem(i2);
            B.AddItem(i3);
            // ジェネリッククラスに要素を(intとして)追加
            gB.AddItem(i1);
            gB.AddItem(i2);
            gB.AddItem(i3);

            Console.WriteLine(B);
            Console.WriteLine(gB);

            // 返り値をキャストします。キャストしない場合はコンパイルエラーが発生します:
            // error CS0266: 型 'object' を 'int' に暗黙的に変換できません...
            iHolder = (int)B.GetItem(1);
            // キャストは必要ありません
            iHolder = gB.GetItem(1);

            string s1 = "s1";
            string s2 = "s2";
            string s3 = "s3";
            string sHolder = "";

            // 標準的クラスに要素を(objectとして)追加
            C.AddItem(s1);
            C.AddItem(s2);
            C.AddItem(s3);
            // int型をstringの配列に追加。全く問題無し。
            C.AddItem(i1);

            // ジェネリッククラスに要素を(stringとして)追加
            gC.AddItem(s1);
            gC.AddItem(s2);
            gC.AddItem(s3);
            // int型をstring型の配列に追加しようとするが、コンパイルエラー発生。
            // error CS1503: 引数 '1': 'int' から 'string' に変換できません。
            //gC.AddItem(i1);

            Console.WriteLine(C);
            Console.WriteLine(gC);

            // 返り値をキャストします。キャストしない場合はコンパイルエラーが発生します:
            // error CS0266: 型 'object' を 'string' に暗黙的に変換できません...
            sHolder = (string)C.GetItem(1);
            // キャストは必要ありません
            sHolder = gC.GetItem(1);
            // stringをintとして取得することはできません
            // error CS0029: 型 'string' を型 'int' に暗黙的に変換できません。
            //iHolder = gC.GetItem(1);
        }

        public class FixedSizeCollection
        {
            /// <summary>
            /// コンストラクタ。静的カウンタをインクリメントするとともに
            /// 要素の最大数を設定します。
            /// </summary>
            /// <param name="maxItems"></param>
            public FixedSizeCollection(int maxItems)
            {
                FixedSizeCollection.InstanceCount++;
                this.Items = new object[maxItems];
            }

            /// <summary>
            /// 要素を追加します。追加される要素の型は不特定のため
            /// object型を引数の型とします。
            /// </summary>
            /// <param name="item">追加する要素</param>
            /// <returns>追加された要素のインデックス</returns>
            public int AddItem(object item)
            {
                if (this.ItemCount < this.Items.Length)
                {
                    this.Items[this.ItemCount] = item;
                    return this.ItemCount++;
                }
                else
                    throw new Exception("要素が最大数に達しています");
            }

            /// <summary>
            /// 要素を取得します。
            /// </summary>
            /// <param name="index">取得する要素のインデックス</param>
            /// <returns>object型の要素</returns>
            public object GetItem(int index)
            {
                if (index >= this.Items.Length &&
                    index >= 0)
                    throw new ArgumentOutOfRangeException("index");

                return this.Items[index];
            }

            #region プロパティ
            /// <summary>
            /// 静的カウンタ。FixedSizeCollectionクラスのインスタンス数を
            /// カウントします。
            /// </summary>
            public static int InstanceCount { get; set; }

            /// <summary>
            /// クラスが保持している要素の数
            /// </summary>
            public int ItemCount { get; private set; }

            /// <summary>
            /// クラスが保持している要素
            /// </summary>
            private object[] Items { get; set; }
            #endregion // プロパティ

            /// <summary>
            /// クラスの詳細情報を表示するためにオーバーライドされたToStringメソッド
            /// </summary>
            /// <returns>クラスの詳細情報を表す整形済みの文字列</returns>
            public override string ToString()
            {
                return this.GetType().ToString() + " のインスタンスが " +
                    FixedSizeCollection.InstanceCount.ToString() +
                    " 個あります。現在のインスタンスが保持している要素は " +
                    this.ItemCount + " 個あります...";
            }
        }

        /// <summary>
        /// インスタンス数を表示可能なジェネリッククラス
        /// </summary>
        /// <typeparam name="T">型引数は配列の要素の型を指定します</typeparam>
        public class FixedSizeCollection<T>
        {
            /// <summary>
            /// コンストラクタ。静的カウンタをインクリメントするとともに、
            /// 内部で保持する要素の最大数を設定します。
            /// </summary>
            /// <param name="items">保持可能な要素の最大数</param>
            public FixedSizeCollection(int items)
            {
                FixedSizeCollection<T>.InstanceCount++;
                this.Items = new T[items];
            }

            /// <summary>
            /// クラスに要素を追加します。
            /// 追加可能な要素の型はインスタンスの生成時に指定された型です。
            /// </summary>
            /// <param name="item">追加する要素</param>
            /// <returns>0を基準とした、追加された要素のインデックス</returns>
            public int AddItem(T item)
            {
                if (this.ItemCount < this.Items.Length)
                {
                    this.Items[this.ItemCount] = item;
                    return this.ItemCount++;
                }
                else
                    throw new Exception("要素が最大数に達しています");
            }

            /// <summary>
            /// クラスに保持された要素を取得します
            /// </summary>
            /// <param name="index">0を基準とした、取得する要素のインデックス</param>
            /// <returns>インスタンスの生成時に指定された型の要素</returns>
            public T GetItem(int index)
            {
                if (index >= this.Items.Length &&
                    index >= 0)
                    throw new ArgumentOutOfRangeException("index");

                return this.Items[index];
            }

            #region プロパティ
            /// <summary>
            /// 静的カウンタ。FixedSizeCollection(Of T)クラスのインスタンス数を
            /// カウントします。
            /// </summary>
            public static int InstanceCount { get; set; }

            /// <summary>
            /// クラスが保持している要素の数
            /// </summary>
            public int ItemCount { get; private set; }

            /// <summary>
            /// クラスが保持している要素
            /// </summary>
            private T[] Items { get; set; }
            #endregion // プロパティ

            /// <summary>
            /// クラスの詳細情報を表示するためにオーバーライドされたToStringメソッド
            /// </summary>
            /// <returns>クラスの詳細情報を表す整形済みの文字列</returns>
            public override string ToString()
            {
                return this.GetType().ToString() + " のインスタンスが " +
                    FixedSizeCollection<T>.InstanceCount.ToString() +
                    " 個あります。現在のインスタンスが保持している要素は " +
                    this.ItemCount + " 個あります...";
            }
        }

        #endregion

        #region "2.3 ジェネリック型のTypeを取得する"
        public static void TestGetGenTypeOf()
        {
            Simple s = new Simple();

            Type t = typeof(Simple);
            Type alsoT = s.GetType();

            // 型パラメータが指定されていないジェネリック型に対応する
            // Typeオブジェクトは存在しないため取得できないことに注意してください。
            // 型パラメータが指定されたジェネリック型に対応する
            // Typeオブジェクトだけが取得可能です。

            // 以下のコードはエラーになります：
            // エラー CS0305: ジェネリック 型 'CSharpRecipes.Generics.SimpleGeneric<T>' の使用には、'1' 型の引数が必要です。
            //Type gt = typeof(SimpleGeneric);

            // 型パラメータを指定したジェネリック型の
            // Typeを取得します
            Type gtInt = typeof(SimpleGeneric<int>);
            Type gtBool = typeof(SimpleGeneric<bool>);
            Type gtString = typeof(SimpleGeneric<string>);

            // 従来の方法と同じく、インスタンスのGetTypeメソッドを呼び出してTypeを
            // 取得することも可能です。
            SimpleGeneric<int> sgI = new SimpleGeneric<int>();
            Type alsoGT = sgI.GetType();
        }

        public class Simple
        {
            public Simple()
            {
            }
        }

        public class SimpleGeneric<T>
        {
            public SimpleGeneric()
            {
            }
        }
        #endregion

        #region "2.4 ArrayListを対応するジェネリッククラスに置き換える"
        public static void UseNonGenericArrayList()
        {
            Console.WriteLine("\r\nUseNonGenericList");

            // ArrayListを作成してデータを格納します
            ArrayList numbers = new ArrayList();
            numbers.Add(1); // ボックス化を発生させます
            numbers.Add(2); // ボックス化を発生させます

            // ArrayList中の全てのint型データを表示します。
            // データそれぞれに対してボックス化解除が発生します。
            foreach (int i in numbers)
            {
                Console.WriteLine(i);
            }

            numbers.Clear();

            Console.WriteLine(numbers.IsReadOnly);
            Console.WriteLine(numbers.IsFixedSize);
            Console.WriteLine(numbers.IsSynchronized);
            Console.WriteLine(numbers.SyncRoot);
        }

        public static void UseGenericList()
        {
            Console.WriteLine("\r\nUseGenericList");

            // Listを生成してデータを格納します
            List<int> numbers = new List<int>();
            numbers.Add(1);
            numbers.Add(2);

            // List中の全てのint型データを表示します。
            foreach (int i in numbers)
            {
                Console.WriteLine(i);
            }

            numbers.Clear();

            Console.WriteLine(((IList<int>)numbers).IsReadOnly);
            Console.WriteLine(((IList)numbers).IsFixedSize);
            Console.WriteLine(((IList)numbers).IsSynchronized);
            Console.WriteLine(((IList)numbers).SyncRoot);
        }

        public static void TestAdapterVSConstructor()
        {
            Console.WriteLine("\r\nAdapter Test");

            ArrayList al = new ArrayList();
            al.Add(new object());
            ArrayList adapter = ArrayList.Adapter(al);
            Console.WriteLine("al " + al[0].GetHashCode());
            Console.WriteLine("adapter " + adapter[0].GetHashCode());

            List<object> oldList = new List<object>();
            oldList.Add(new object());
            List<object> newList = new List<object>(oldList);
            Console.WriteLine("oldList " + oldList[0].GetHashCode());
            Console.WriteLine("newList " + newList[0].GetHashCode());
        }

        public static void TestCloneVSGetRange()
        {
            Console.WriteLine("\r\nClone Test");

            ArrayList al = new ArrayList();
            al.Add(new object());
            ArrayList clone = (ArrayList)al.Clone();
            Console.WriteLine("al " + al[0].GetHashCode());
            Console.WriteLine("clone " + clone[0].GetHashCode());

            List<object> oldList = new List<object>();
            oldList.Add(new object());
            List<object> newList = oldList.GetRange(0, oldList.Count);
            Console.WriteLine("oldList " + oldList[0].GetHashCode());
            Console.WriteLine("newList " + newList[0].GetHashCode());
        }

        public static void TestGenericRepeat()
        {
            List<int> numberList = new List<int>();
            numberList.Repeat(100, 3);

            foreach (int i in numberList)
                Console.WriteLine(i);
        }

        public static void Repeat<T>(this List<T> list, T obj, int count)
        {
            if (count < 0)
            {
                throw (new ArgumentException("引数 count には0以上を指定する必要があります。"));
            }

            for (int index = 0; index < count; index++)
            {
                list.Add(obj);
            }
        }

        public static void CloneGenericList()
        {
            Console.WriteLine("\r\nCloneGenericList");

            List<int> numbers = new List<int>();
            numbers.Add(1);
            numbers.Add(2);

            List<int> cloneNumbers = new List<int>(numbers);
            Console.WriteLine("cloneNumbers[0] " + cloneNumbers[0].GetHashCode());

            Console.WriteLine(((IList<int>)numbers).IsReadOnly);
            Console.WriteLine(((IList)numbers).IsFixedSize);
            Console.WriteLine(((IList)numbers).IsSynchronized);
            Console.WriteLine(((IList)numbers).SyncRoot);
        }
        #endregion

        #region "4.5 StackとQueueを対応するジェネリッククラスに置き換える"

        public static void UseGenericStack()
        {
            // ジェネリックStackオブジェクトを作成します
            Stack<int> numericStack = new Stack<int>();

            // Stackにデータを格納します
            numericStack.Push(1);
            numericStack.Push(2);
            numericStack.Push(3);

            // Stackからデータを取り出して表示します
            Console.WriteLine(numericStack.Pop().ToString());
            Console.WriteLine(numericStack.Pop().ToString());
            Console.WriteLine(numericStack.Pop().ToString());
        }

        public static void UseNonGenericStack()
        {
            // 非ジェネリックStackオブジェクトを作成します
            Stack numericStack = new Stack();

            // Stackにデータを格納します(ボックス化が発生します)
            numericStack.Push(1);
            numericStack.Push(2);
            numericStack.Push(3);

            // Stackからデータを取り出して表示します(ボックス化解除が発生します)
            Console.WriteLine(numericStack.Pop().ToString());
            Console.WriteLine(numericStack.Pop().ToString());
            Console.WriteLine(numericStack.Pop().ToString());
        }

        public static void CloneStack()
        {
            // ジェネリックStackオブジェクトを作成します
            Stack<int> numericStack = new Stack<int>();

            // Stackにデータを格納します
            numericStack.Push(1);
            numericStack.Push(2);
            numericStack.Push(3);

            // numericStackの複製を作成します
            Stack<int> clonedNumericStack = new Stack<int>(numericStack);

            // Stackからpopせずにpeekで値を参照します
            foreach (int i in clonedNumericStack)
            {
                Console.WriteLine("foreach: " + i.ToString());
            }

            // Stackからpopして値を取り出しつつ表示します
            Console.WriteLine(clonedNumericStack.Pop().ToString());
            Console.WriteLine(clonedNumericStack.Pop().ToString());
            Console.WriteLine(clonedNumericStack.Pop().ToString());
        }


        public static void UseGenericQueue()
        {
            // ジェネリックQueueオブジェクトを作成します
            Queue<int> numericQueue = new Queue<int>();

            // Queueにデータを格納します
            numericQueue.Enqueue(1);
            numericQueue.Enqueue(2);
            numericQueue.Enqueue(3);

            // Queueからデータを取り出して表示します
            Console.WriteLine(numericQueue.Dequeue());
            Console.WriteLine(numericQueue.Dequeue());
            Console.WriteLine(numericQueue.Dequeue());
        }

        public static void UseNonGenericQueue()
        {
            // 非ジェネリックQueueオブジェクトを作成します
            Queue numericQueue = new Queue();

            // Queueにデータを格納します(ボックス化が行われることになります)
            numericQueue.Enqueue(1);
            numericQueue.Enqueue(2);
            numericQueue.Enqueue(3);

            // Queueからデータを取り出して表示します(ボックス化解除が行われることになります)
            Console.WriteLine(numericQueue.Dequeue());
            Console.WriteLine(numericQueue.Dequeue());
            Console.WriteLine(numericQueue.Dequeue());
        }

        public static void CloneQueue()
        {
            // ジェネリックQueueオブジェクトを作成します
            Queue<int> numericQueue = new Queue<int>();

            // Queueにデータを格納します
            numericQueue.Enqueue(1);
            numericQueue.Enqueue(2);
            numericQueue.Enqueue(3);

            // numericQueueの複製を作成します
            Queue<int> clonedNumericQueue = new Queue<int>(numericQueue);

            // Queueからpopせずにpeekで値を参照します
            foreach (int i in clonedNumericQueue)
            {
                Console.WriteLine("foreach: " + i.ToString());
            }

            // Queueからpopして値を取り出しつつ表示します
            Console.WriteLine(clonedNumericQueue.Dequeue());
            Console.WriteLine(clonedNumericQueue.Dequeue());
            Console.WriteLine(clonedNumericQueue.Dequeue());
        }
        #endregion

        #region "2.6 LinkedListを使用する"
        public static void UseLinkedList()
        {
            // LinkedListに追加するためのToDoItemを作成します
            ToDoItem i1 =
                new ToDoItem() { Name = "ドアにペンキを塗る", Comment = "3番目にすること" };
            ToDoItem i2 =
                new ToDoItem() { Name = "ドアを買う", Comment = "最初にすること" };
            ToDoItem i3 =
                new ToDoItem() { Name = "ドアを組み立てる", Comment = "2番目にすること" };
            ToDoItem i4 =
                new ToDoItem() { Name = "ドアを取り付ける", Comment = "最後にすること" };

            // 新しいLinkedListオブジェクトを作成します
            LinkedList<ToDoItem> todoList = new LinkedList<ToDoItem>();

            // Add the items
            todoList.AddFirst(i1);
            todoList.AddFirst(i2);
            todoList.AddBefore(todoList.Find(i1), i3);
            todoList.AddAfter(todoList.Find(i1), i4);

            // すべての要素を表示します
            foreach (ToDoItem tdi in todoList)
            {
                Console.WriteLine(tdi.Name + " : " + tdi.Comment);
            }

            // LinkedListの1番目の要素についての情報を表示します
            Console.WriteLine("todoList.First.Value.Name == " +
                todoList.First.Value.Name);

            // LinkedListの1番目の要素についての情報を表示します
            Console.WriteLine("todoList.First.Next.Value.Name == " +
                todoList.First.Next.Value.Name);

            // LinkedListの最後から1つ前の要素についての情報を表示します
            Console.WriteLine("todoList.Last.Previous.Value.Name == " +
                todoList.Last.Previous.Value.Name);
        }

        /// <summary>
        /// Todo リストの項目
        /// </summary>
        public class ToDoItem
        {
            /// <summary>
            /// 項目名
            /// </summary>
            public string Name { get; set; }

            /// <summary>
            /// 項目に関するコメント
            /// </summary>
            public string Comment { get; set; }
        }
        #endregion

        #region "2.7 初期値nullの値型を作成する"
        public static void TestNullableStruct()
        {
            Console.WriteLine("\r\n\r\n");

            //int? myDBInt = null;
            // または
            Nullable<int> myDBInt = new Nullable<int>();
            Nullable<int> myTempDBInt = new Nullable<int>();


            if (myDBInt.HasValue)
                Console.WriteLine("値有り:  " + myDBInt.Value);
            else
                Console.WriteLine("値がありません(NULL)");

            myDBInt = 100;

            if (myDBInt != null)
                Console.WriteLine("値有り:  " + myDBInt.Value);
            else
                Console.WriteLine("値がありません(NULL)");

            if (myTempDBInt != null)
            {
                if (myTempDBInt < 100)
                    Console.WriteLine("myTempDBInt < 100");
                else
                    Console.WriteLine("myTempDBInt >= 100");
            }
            else
            {
                // nullの場合の処理をここに記述します
            }
        }
        #endregion

        #region "2.8 SortedListの要素を逆順にする"

        public static void TestReversibleSortedList()
        {
            SortedList<int, string> data = new SortedList<int, string>();
            data.Add(2, "two");
            data.Add(5, "five");
            data.Add(3, "three");
            data.Add(1, "one");

            foreach (KeyValuePair<int, string> kvp in data)
            {
                Debug.WriteLine("\t" + kvp.Key + "\t" + kvp.Value);
            }
            Debug.WriteLine("");

            // 逆順で要素を取得します
            var query = from d in data
                        orderby d.Key descending
                        select d;

            foreach (KeyValuePair<int, string> kvp in query)
            {
                Debug.WriteLine("\t" + kvp.Key + "\t" + kvp.Value);
            }
            Debug.WriteLine("");


            data.Add(4, "four");

            // 再び逆順で要素を取得します
            query = from d in data
                    orderby d.Key descending
                    select d;

            foreach (KeyValuePair<int, string> kvp in query)
            {
                Debug.WriteLine("\t" + kvp.Key + "\t" + kvp.Value);
            }
            Debug.WriteLine("");

            // さらに通常の順序で要素を追加します
            foreach (KeyValuePair<int, string> kvp in data)
            {
                Debug.WriteLine("\t" + kvp.Key + "\t" + kvp.Value);
            }
        }
        #endregion

        #region "2.9 ジェネリック型の読み取り専用コレクションの作成"
        public static void MakeCollectionReadOnly()
        {
            Lottery tryYourLuck = new Lottery();
            // 結果を表示します
            for (int i = 0; i < tryYourLuck.Results.Count; i++)
            {
                Console.WriteLine("ロト番号 " + i + " は " + tryYourLuck.Results[i]);
            }

            int[] items = { 0, 1, 2 };
            ReadOnlyCollection<int> readOnlyItems =
                new ReadOnlyCollection<int>(items);

            // 番号を当選するように変更！
            //tryYourLuck.Results[0] = 29;

            // 上の行のコメントを外すと次のコンパイルエラーが発生します：
            // エラー CS0200: プロパティまたはインデクサ 'System.Collections.ObjectModel.ReadOnly
            // Collection<int>.this[int]' は読み取り専用なので、割り当てることはできません。
            // Error    26
        }

        public class Lottery
        {
            // リストを作成
            List<int> _numbers;

            public Lottery()
            {
                // 当選番号を設定
                _numbers = new List<int>(5) { 17, 21, 32, 44, 58 };
            }

            public ReadOnlyCollection<int> Results
            {
                // 結果のコピーをラップして返します
                get { return new ReadOnlyCollection<int>(_numbers); }
            }
        }

        #endregion

        #region "2.10 Hashtableを対応するジェネリッククラスに置き換える"
        public static void UseNonGenericHashtable()
        {
            Console.WriteLine("\r\nUseNonGenericHashtable");

            // Hashtableを作成してデータを格納します
            var numbers = new Hashtable() 
                { {1, "one"}, // キーに対してボックス化が発生します
                  {2, "two"} }; // キーに対してボックス化が発生します

            // Hashtableのすべての値を表示します。
            // キーの値それぞれについてボックス化解除が発生します。
            foreach (DictionaryEntry de in numbers)
            {
                Console.WriteLine("Key: " + de.Key + "\tValue: " + de.Value);
            }

            Console.WriteLine(numbers.IsReadOnly);
            Console.WriteLine(numbers.IsFixedSize);
            Console.WriteLine(numbers.IsSynchronized);
            Console.WriteLine(numbers.SyncRoot);

            numbers.Clear();
        }

        public static void UseGenericDictionary()
        {
            Console.WriteLine("\r\nUseGenericDictionary");

            // Dictionaryを作成してデータを格納します
            Dictionary<int, string> numbers =
                new Dictionary<int, string>() { { 1, "one" }, { 2, "two" } };

            // Dictionaryのすべてのキー/値のペアを表示します
            foreach (KeyValuePair<int, string> kvp in numbers)
            {
                Console.WriteLine("Key: " + kvp.Key + "\tValue: " + kvp.Value);
            }

            Console.WriteLine(((IDictionary)numbers).IsReadOnly);
            Console.WriteLine(((IDictionary)numbers).IsFixedSize);
            Console.WriteLine(((IDictionary)numbers).IsSynchronized);
            Console.WriteLine(((IDictionary)numbers).SyncRoot);

            numbers.Clear();
        }

        public static void CopyToGenericDictionary()
        {
            Console.WriteLine("\r\nCopyToGenericDictionary");

            // Dictionaryを作成してデータを格納します
            Dictionary<int, string> numbers =
                new Dictionary<int, string>() { { 1, "one" }, { 2, "two" } };

            // Dictionaryのすべてのキー/値のペアを表示します
            foreach (KeyValuePair<int, string> kvp in numbers)
            {
                Console.WriteLine("Key: " + kvp.Key + "\tValue: " + kvp.Value);
            }

            // Dictionaryクラスの複製データを保持するための配列を作成します
            KeyValuePair<int, string>[] objs = new KeyValuePair<int, string>[numbers.Count];

            // DictionaryのCopyToメソッドを呼び出して
            // Dictionaryオブジェクトに格納されていたすべてのKeyValuePairをobjsにコピーします
            ((IDictionary)numbers).CopyTo(objs, 0);

            // objsのすべてのキー/値のペアを表示します
            foreach (KeyValuePair<int, string> kvp in objs)
            {
                Console.WriteLine("Key: " + kvp.Key + "\tValue: " + kvp.Value);
            }
        }

        public static void CloneGenericDictionary()
        {
            Console.WriteLine("\r\nCloneGenericDictionary");

            // Dictionaryを作成してデータを格納します
            Dictionary<int, string> numbers =
                new Dictionary<int, string>() { { 1, "one" }, { 2, "two" } };

            // Dictionaryのすべてのキー/値のペアを表示します
            foreach (KeyValuePair<int, string> kvp in numbers)
            {
                Console.WriteLine("元データのKey: " + kvp.Key + "\tValue: " + kvp.Value);
            }

            // Dictionaryオブジェクトを複製します
            Dictionary<int, string> clonedNumbers =
                new Dictionary<int, string>(numbers);

            // 複製されたDictionaryオブジェクトのすべてのキー/値のペアを表示します
            foreach (KeyValuePair<int, string> kvp in numbers)
            {
                Console.WriteLine("複製データのKey: " + kvp.Key + "\tValue: " + kvp.Value);
            }
        }
        #endregion

        #region "2.11 Dictionaryジェネリッククラスに対してforeachを使用する"
        public static void ShowForEachWithDictionary()
        {
            // Dictionaryオブジェクトを作成してデータを格納します
            Dictionary<int, string> myStringDict =
                new Dictionary<int, string>() { { 1, "Foo" }, { 2, "Bar" }, { 3, "Baz" } };

            // すべてのキー/値のペアを走査して表示します
            foreach (KeyValuePair<int, string> kvp in myStringDict)
            {
                Console.WriteLine("key   " + kvp.Key);
                Console.WriteLine("Value " + kvp.Value);
                Console.WriteLine("kvp " + kvp.ToString());
            }

            // Hashtableオブジェクトに対してはDictionaryEntryを使用します。
            // DictionaryオブジェクトにDictionaryEntryを使用できないことに注意してください。
            Hashtable myHashtable =
                new Hashtable() { { 1, "Foo" }, { 2, "Bar" }, { 3, "Baz" } };
            foreach (DictionaryEntry de in myHashtable)
            {
                Console.WriteLine("key   " + de.Key);
                Console.WriteLine("Value " + de.Value);
                Console.WriteLine("kvp " + de.ToString());
            }
        }
        #endregion

        #region "2.12 型パラメータに制限をかける"
        public static void TestConversionCls()
        {
            Console.WriteLine("\r\n\r\n");

            Conversion<long> c = new Conversion<long>();
            //Console.WriteLine("long.MinValue:  " + c.ShowAsInt(long.MinValue));
            Console.WriteLine("-100:  " + c.ShowAsInt(-100));
            Console.WriteLine("0:  " + c.ShowAsInt(0));
            Console.WriteLine("100:  " + c.ShowAsInt(100));
            //Console.WriteLine("long.MaxValue:  " + c.ShowAsInt(long.MaxValue));
        }

        public static void TestComparableListCls()
        {
            Console.WriteLine("\r\n\r\n");

            ComparableList<int> cp =
                new ComparableList<int>() { 100, 10 };

            Console.WriteLine("0 compare 1 == " + cp.Compare(0, 1));
            Console.WriteLine("1 compare 0 == " + cp.Compare(1, 0));
            Console.WriteLine("1 compare 1  == " + cp.Compare(1, 1));
        }

        public static void TestDisposableListCls()
        {
            Console.WriteLine("\r\n\r\n");

            DisposableList<StreamReader> dl = new DisposableList<StreamReader>();

            // Create a few test objects
            StreamReader tr1 = new StreamReader("../../00_MainTester.cs");
            StreamReader tr2 = new StreamReader("../../01_LINQ.cs");
            StreamReader tr3 = new StreamReader("../../02_Generics.cs");

            // Add the test object to the DisposableList
            dl.Add(tr1);
            dl.Insert(0, tr2);

            Console.WriteLine("dl.IndexOf(tr3) == " + dl.IndexOf(tr3));

            dl.Add(tr3);

            Console.WriteLine("dl.Contains(tr1) == " + dl.Contains(tr1));

            StreamReader[] srArray = new StreamReader[3];
            dl.CopyTo(srArray, 0);
            Console.WriteLine("srArray[1].ReadLine() == " + srArray[1].ReadLine());

            Console.WriteLine("dl.Count == " + dl.Count);

            foreach (StreamReader sr in dl)
            {
                Console.WriteLine("sr.ReadLine() == " + sr.ReadLine());
            }

            Console.WriteLine("dl.IndexOf(tr3) == " + dl.IndexOf(tr3));

            Console.WriteLine("dl.IsReadOnly == " + dl.IsReadOnly);

            // Call Dispose before any of the disposable objects are removed from the DisposableList
            dl.RemoveAt(0);
            Console.WriteLine("dl.Count == " + dl.Count);

            dl.Remove(tr1);
            Console.WriteLine("dl.Count == " + dl.Count);

            dl.Clear();
            Console.WriteLine("dl.Count == " + dl.Count);
        }


        public class Conversion<T>
            where T : struct, IConvertible
        {
            public int ShowAsInt(T value)
            {
                return (value.ToInt32(NumberFormatInfo.CurrentInfo));
            }
        }

        public class ComparableList<T> : List<T>
            where T : IComparable<T>
        {
            public int Compare(int index1, int index2)
            {
                return (index1.CompareTo(index2));
            }
        }

        public class DisposableList<T> : IList<T>
            where T : class, IDisposable
        {
            private List<T> _items = new List<T>();

            // リスト中の要素を破棄するためのprivateメソッド
            private void Delete(T item)
            {
                item.Dispose();
            }

            // IList<T> メンバ
            public int IndexOf(T item)
            {
                return (_items.IndexOf(item));
            }

            public void Insert(int index, T item)
            {
                _items.Insert(index, item);
            }

            public T this[int index]
            {
                get { return (_items[index]); }
                set { _items[index] = value; }
            }

            public void RemoveAt(int index)
            {
                Delete(this[index]);
                _items.RemoveAt(index);
            }

            // ICollection<T> メンバ
            public void Add(T item)
            {
                _items.Add(item);
            }

            public bool Contains(T item)
            {
                return (_items.Contains(item));
            }

            public void CopyTo(T[] array, int arrayIndex)
            {
                _items.CopyTo(array, arrayIndex);
            }

            public int Count
            {
                get { return (_items.Count); }
            }

            public bool IsReadOnly
            {
                get { return (false); }
            }

            // IEnumerable<T> メンバ
            public IEnumerator<T> GetEnumerator()
            {
                return (_items.GetEnumerator());
            }

            // IEnumerable メンバ
            IEnumerator IEnumerable.GetEnumerator()
            {
                return (_items.GetEnumerator());
            }

            // その他のメンバ
            public void Clear()
            {
                for (int index = 0; index < _items.Count; index++)
                {
                    Delete(_items[index]);
                }

                _items.Clear();
            }

            public bool Remove(T item)
            {
                int index = _items.IndexOf(item);

                if (index >= 0)
                {
                    Delete(_items[index]);
                    _items.RemoveAt(index);

                    return (true);
                }
                else
                {
                    return (false);
                }
            }
        }
        #endregion

        #region "2.13 ジェネリック型の特定の変数を型の既定値で初期化する"
        public static void ShowSettingFieldsToDefaults()
        {
            Console.WriteLine("\r\n\r\n");

            DefaultValueExample<int> dv = new DefaultValueExample<int>();

            // データが初期値に設定されているかどうか確認します。
            // 設定されている場合にはtrueが返されます。
            bool isDefault = dv.IsDefaultData();
            Console.WriteLine("初期値: " + isDefault);

            // データを設定します
            dv.SetData(100);

            // 再び確認します。今度はfalseが返されます
            isDefault = dv.IsDefaultData();
            Console.WriteLine("データ設定後: " + isDefault);
        }

        public class DefaultValueExample<T>
        {
            T data = default(T);

            public bool IsDefaultData()
            {
                T temp = default(T);

                return temp.Equals(data);
            }

            public void SetData(T value)
            {
                data = value;
            }
        }
        #endregion

    }
}
