﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

namespace CSharpRecipes
{
    public static class IteratorExtensionMethods
    {
        public static IEnumerable<T> EveryNthItem<T>(this IEnumerable<T> enumerable, int step)
        {
            int current = 0;
            foreach (T item in enumerable)
            {
                ++current;
                if (current % step == 0)
                    yield return item;
            }
        }
    }

    public partial class IteratorsAndPartialTypesAndPartialMethods
    {
        #region "3.1 ジェネリック型の反復子を作成する"
        public static void TestShoppingCart()
        {
            // ShoppingListオブジェクトを作成してデータを格納します
            ShoppingList<string> shoppingCart = new ShoppingList<string>(){
                "item1","item2","item3","item4","item5","item6"};

            // ShoppingListオブジェクトに格納された全ての要素を出力します
            foreach (string item in shoppingCart)
            {
                Console.WriteLine(item);
            }
        }


        public class ShoppingList<T> : IEnumerable<T>
        {
            private List<T> _items = new List<T>();

            public void Add(T name)
            {
                _items.Add(name);
            }

            public IEnumerator<T> GetEnumerator()
            {
                return _items.GetEnumerator();
            }

            IEnumerator<T> IEnumerable<T>.GetEnumerator()
            {
                return GetEnumerator();
            }

            IEnumerator IEnumerable.GetEnumerator()
            {
                return GetEnumerator();
            }
        }
        #endregion

        #region "3.2 非ジェネリック型の反復子を作成する"
        public static void TestStampCollection()
        {
            // StampCollectionオブジェクトを作成してデータを格納します
            StampCollection stamps = new StampCollection() {
                new Stamp(1998,"Louisiana Duck"),
                new Stamp(1968,"Goethals Memorial"),
                new Stamp(1909,"Carmine Hudson"),
                new Stamp(1936,"Hotel Corner Card")};

            foreach (Stamp stamp in stamps)
            {
                Console.WriteLine(stamp);
            }
        }

        public class Stamp
        {
            public Stamp(int year, string name)
            {
                this.Year = year;
                this.Name = name;
            }
            public int Year { get; set; }
            public string Name { get; set; }

            public override string ToString()
            {
                return this.Year + ":" + this.Name;
            }
        }

        public class StampCollection : IEnumerable
        {
            private Dictionary<string, Stamp> _stamps =
                new Dictionary<string, Stamp>();

            public void Add(Stamp stamp)
            {
                _stamps.Add(stamp.Name, stamp);
            }

            public IEnumerator GetEnumerator()
            {
                // このオブジェクトに保持されたすべての切手を
                // 発行日順に返します
                var orderedStamps = from Stamp stamp in _stamps.Values
                                    orderby stamp.Year
                                    select stamp;
                foreach (Stamp stamp in orderedStamps)
                {
                    yield return stamp;
                }
            }
        }
        #endregion

        #region "3.3 独自の列挙子を作成する"
        public static void TestIterators()
        {
            Container<int> container = new Container<int>();

            // テストデータを作成します
            List<int> testData = new List<int>(){
	            -1,1,2,3,4,5,6,7,8,9,10,200,500};

            // Add test data to Container object
            container.Clear();
            container.AddRange(testData);

            // Containerオブジェクトを列挙します
            foreach (int i in container)
            {
                Console.WriteLine(i);
            }

            Console.WriteLine();
            foreach (int i in container.GetReverseOrderEnumerator())
            {
                Console.WriteLine(i);
            }

            Console.WriteLine();
            foreach (int i in container.GetForwardStepEnumerator(2))
            {
                Console.WriteLine(i);
            }

            Console.WriteLine();
            foreach (int i in container.GetReverseStepEnumerator(3))
            {
                Console.WriteLine(i);
            }
        }

        public class Container<T> : IEnumerable<T>
        {
            public Container() { }

            private List<T> _internalList = new List<T>();

            // 要素を先頭から末尾の順で列挙します
            public IEnumerator<T> GetEnumerator()
            {
                return _internalList.GetEnumerator();
            }

            // 要素を末尾から先頭の順で列挙します
            public IEnumerable<T> GetReverseOrderEnumerator()
            {
                foreach (T item in ((IEnumerable<T>)_internalList).Reverse())
                {
                    yield return item;
                }
            }

            // 要素を特定の数だけとばして先頭から末尾の順で列挙します
            public IEnumerable<T> GetForwardStepEnumerator(int step)
            {
                foreach (T item in _internalList.EveryNthItem(step))
                {
                    yield return item;
                }
            }

            // 要素を特定の数だけとばして末尾から先頭の順で列挙します
            public IEnumerable<T> GetReverseStepEnumerator(int step)
            {
                foreach (T item in ((IEnumerable<T>)_internalList).Reverse().EveryNthItem(step))
                {
                    yield return item;
                }
            }

            #region IEnumerable メンバ

            IEnumerator IEnumerable.GetEnumerator()
            {
                return GetEnumerator();
            }

            #endregion

            public void Clear()
            {
                _internalList.Clear();
            }

            public void Add(T item)
            {
                _internalList.Add(item);
            }

            public void AddRange(ICollection<T> collection)
            {
                _internalList.AddRange(collection);
            }
        }


        #endregion

        #region "3.4 反復子のロジックを実装する"
        public static void TestIteratorsAndLinq()
        {
            // SectionalListを作成してデータを格納します
            SectionalList<int> sectionalList = new SectionalList<int>() {
                12,26,95,37,50,33,81,54};

            // SectionalListの全データを表示します
            Console.WriteLine("\r\nGetEnumerator反復子");
            foreach (int i in sectionalList)
            {
                Console.Write(i + ":");
            }
            Console.WriteLine("");

            Console.WriteLine("\r\nGetFirstHalf反復子");
            foreach (int i in sectionalList.GetFirstHalf())
            {
                Console.Write(i + ":");
            }
            Console.WriteLine("");

            Console.WriteLine("\r\nGetSecondHalf反復子");
            foreach (int i in sectionalList.GetSecondHalf())
            {
                Console.Write(i + ":");
            }
            Console.WriteLine("");

            Console.WriteLine("\r\nGetFilteredValues反復子");
            // 偶数を基準として要素を取得します
            Func<int, bool> predicate = item => (item % 2 == 0);
            foreach (int i in sectionalList.GetFilteredValues(predicate))
            {
                Console.Write(i + ":");
            }
            Console.WriteLine("");

            Console.WriteLine("\r\nGetNonFilteredValues反復子");
            foreach (int i in sectionalList.GetNonFilteredValues(predicate))
            {
                Console.Write(i + ":");
            }
            Console.WriteLine("");

            Console.WriteLine("\r\nGetReverseFilteredValues反復子");
            foreach (int i in sectionalList.GetReverseFilteredValues())
            {
                Console.Write(i + ":");
            }
            Console.WriteLine("");
        }


        public class SectionalList<T> : IEnumerable<T>
        {
            private List<T> _items = new List<T>();

            public void Add(T item)
            {
                _items.Add(item);
            }

            public IEnumerator<T> GetEnumerator()
            {
                return _items.GetEnumerator();
            }

            IEnumerator IEnumerable.GetEnumerator()
            {
                return GetEnumerator();
            }

            public IEnumerable<T> GetFirstHalf()
            {
                foreach (T item in _items.Take(_items.Count / 2))
                {
                    yield return item;
                }
            }
            public IEnumerable<T> GetSecondHalf()
            {
                foreach (T item in _items.Skip(_items.Count / 2))
                {
                    yield return item;
                }
            }

            public IEnumerable<T> GetFilteredValues(Func<T, bool> predicate)
            {
                foreach (T item in _items.TakeWhile(predicate))
                {
                    yield return item;
                }
            }

            public IEnumerable<T> GetNonFilteredValues(Func<T, bool> predicate)
            {
                foreach (T item in _items.SkipWhile(predicate))
                {
                    yield return item;
                }
            }

            public IEnumerable<T> GetReverseFilteredValues()
            {
                foreach (T item in ((IEnumerable<T>)_items).Reverse())
                {
                    yield return item;
                }
            }
        }
        #endregion

        #region "3.5 反復処理を強制的に終了する"
        public static void TestYieldBreak()
        {
            // UpperLimitListを作成してデータを格納します
            UpperLimitList<string> gatePasses = new UpperLimitList<string>() {
                "A","B","C","D","E","F","G"};

            Console.WriteLine("上限を設定する前のゲート");
            foreach (string gatePass in gatePasses)
            {
                Console.Write(gatePass + ":");
            }
            Console.WriteLine("");

            // ゲートの通過上限を5に設定します
            gatePasses.UpperLimit = 5;

            Console.WriteLine("上限を設定後のゲート");
            foreach (string gatePass in gatePasses)
            {
                Console.Write(gatePass + ":");
            }
            Console.WriteLine("");
        }


        public class UpperLimitList<T> : IEnumerable<T>
        {
            private List<T> _items = new List<T>();
            private bool noMoreItemsCanBeAdded;
            private int upperLimit = -1;

            public int UpperLimit
            {
                get { return upperLimit; }
                set { upperLimit = value; }
            }

            public void Add(T name)
            {
                _items.Add(name);
            }

            public IEnumerator<T> GetEnumerator()
            {
                for (int index = 0; index < _items.Count; index++)
                {
                    if (noMoreItemsCanBeAdded)
                    {
                        yield break;
                    }
                    else
                    {
                        // noMoreItemsCanBeAddedをtrueにするような判定を行います
                        if (upperLimit >= 0 && index >= upperLimit - 1)
                        {
                            noMoreItemsCanBeAdded = true;
                        }

                        yield return _items[index];
                    }
                }
            }

            IEnumerator IEnumerable.GetEnumerator()
            {
                return GetEnumerator();
            }
        }
        #endregion

        #region "3.6 反復子とfinallyブロックを組み合わせる"
        public static void TestFinallyAndIterators()
        {
            // StringSet オブジェクトを作成してデータを格納します
            StringSet strSet =
                new StringSet()
                    {"item1", 
                     "item2", 
                     "item3", 
                     "item4", 
                     "item5"};

            //// GetEnumerator反復子を使用します
            //foreach (string s in strSet)
            //{
            //    Console.WriteLine(s);
            //}


            // StringSet オブジェクトの全ての要素を表示します
            try
            {
                int i = 0;
                foreach (string s in strSet)
                {
                    try
                    {
                        Console.WriteLine(s);
                        // ここで強制的に例外を発生させます
                        if (i++ == 1) throw new Exception();
                    }
                    catch (Exception)
                    {
                        Console.WriteLine("foreachのcatchブロックの中です");
                    }
                    finally
                    {
                        // 反復の度に実行されます
                        Console.WriteLine("foreachのfinallyブロックの中です");
                    }
                }
            }
            catch (Exception)
            {
                Console.WriteLine("外側のcatchブロックの中です");
            }
            finally
            {
                Console.WriteLine("外側のfinallyブロックの中です");
            }

            /*
            反復処理中に例外が発生した場合の出力は以下の通りです：
             - In iterator finally block
             - In outer catch block
             - In outer finally block

            反復処理中に例外が発生しなかった場合の出力は以下の通りです：
             - item1
             - In foreach finally block
             - item2
             - In foreach finally block
             - item3
             - In foreach finally block
             - item4
             - In foreach finally block
             - item5
             - In foreach finally block
             - item6
             - In foreach finally block
             - In iterator finally block
             - In outer finally block

            foreachループ中で例外が発生した場合の出力は以下の通りです：
             - In foreach catch block
             - In foreach finally block
             - In foreach catch block
             - In foreach finally block
             - In foreach catch block
             - In foreach finally block
             - In foreach catch block
             - In foreach finally block
             - In foreach catch block
             - In foreach finally block
             - In foreach catch block
             - In foreach finally block
             - In iterator finally block
             - In outer finally block
            */
        }


        public class StringSet : IEnumerable<string>
        {
            private List<string> _items = new List<string>();

            public void Add(string value)
            {
                _items.Add(value);
            }

            public IEnumerator<string> GetEnumerator()
            {
                try
                {
                    for (int index = 0; index < _items.Count; index++)
                    {
                        // ここで強制的に例外を発生させます
                        //if (index == 1) throw new Exception();
                        yield return _items[index];
                    }
                }
                // 反復子中ではcatchブロックを使用できません
                finally
                {
                    // foreachループが終了した時点でのみ実行されます(yield breakが呼び出された場合も同様です)
                    Console.WriteLine("反復子のfinallyブロックの中です");
                }
            }

            #region IEnumerable Members

            IEnumerator IEnumerable.GetEnumerator()
            {
                return GetEnumerator();
            }

            #endregion
        }
        #endregion

        #region "3.7 クラスがネストされたforeachをサポートできるようにする"
        public static void CreateNestedObjects()
        {
            Group<Group<Item>> hierarchy =
                new Group<Group<Item>>("root") {
                    new Group<Item>("subgroup1"){
                        new Item("item1",100),
                        new Item("item2",200)},
                    new Group<Item>("subgroup2"){
                        new Item("item3",300),
                        new Item("item4",400)}};

            IEnumerator enumerator = ((IEnumerable)hierarchy).GetEnumerator();
            while (enumerator.MoveNext())
            {
                Console.WriteLine(((Group<Item>)enumerator.Current).Name);
                foreach (Item i in ((Group<Item>)enumerator.Current))
                {
                    Console.WriteLine(i.Name);
                }
            }

            // データを読み取ります
            DisplayNestedObjects(hierarchy);
        }

        //topLevelGroup.Count: 2
        //topLevelGroupName:  root
        //        subGroup.SubGroupName:  subgroup1
        //        subGroup.Count: 2
        //                item.Name:     item1
        //                item.Location: 100
        //                item.Name:     item2
        //                item.Location: 200
        //        subGroup.SubGroupName:  subgroup2
        //        subGroup.Count: 2
        //                item.Name:     item3
        //                item.Location: 300
        //                item.Name:     item4
        //                item.Location: 400

        private static void DisplayNestedObjects(Group<Group<Item>> topLevelGroup)
        {
            Console.WriteLine("topLevelGroup.Count: " + topLevelGroup.Count);
            Console.WriteLine("topLevelGroupName:  " + topLevelGroup.Name);

            // 外側のforeachでは1段目のオブジェクトを走査します
            foreach (Group<Item> subGroup in topLevelGroup)
            {
                Console.WriteLine("\tsubGroup.SubGroupName:  " + subGroup.Name);
                Console.WriteLine("\tsubGroup.Count: " + subGroup.Count);

                // 内側のforeachでは現在の子Groupオブジェクトに保持された
                // Itemオブジェクトからデータを読み取ります
                foreach (Item item in subGroup)
                {
                    Console.WriteLine("\t\titem.Name:     " + item.Name);
                    Console.WriteLine("\t\titem.Location: " + item.Location);
                }
            }
        }


        public class Group<T> : IEnumerable<T>
        {
            public Group(string name)
            {
                this.Name = name;
            }

            private List<T> _groupList = new List<T>();

            public string Name { get; set; }

            public int Count
            {
                get { return _groupList.Count; }
            }

            public void Add(T group)
            {
                _groupList.Add(group);
            }

            IEnumerator IEnumerable.GetEnumerator()
            {
                return GetEnumerator();
                //return new GroupEnumerator<T>(_groupList.ToArray());
            }

            public IEnumerator<T> GetEnumerator()
            {
                return _groupList.GetEnumerator();
            }
        }


        public class Item
        {
            public Item(string name, int location)
            {
                this.Name = name;
                this.Location = location;
            }
            public string Name { get; set; }
            public int Location { get; set; }
        }

        public class GroupEnumerator<T> : IEnumerator
        {
            public T[] _items;

            int position = -1;

            public GroupEnumerator(T[] list)
            {
                _items = list;
            }

            public bool MoveNext()
            {
                position++;
                return position < _items.Length;
            }

            public void Reset()
            {
                position = -1;
            }

            public object Current
            {
                get
                {
                    try
                    {
                        return _items[position];
                    }
                    catch (IndexOutOfRangeException)
                    {
                        throw new InvalidOperationException();
                    }
                }
            }
        }


        #endregion

        #region "3.8 インターフェイスの実装を整頓する"
        // "PartialClassInterface"プロジェクトを参照してください
        #endregion

        #region "3.9 メインのコードパスから外れたコードを自動生成する"
        // "PartialClassAddin"プロジェクトを参照してください
        #endregion

        #region "3.10 エンティティの生成時に行う処理を追加する"
        public static void TestPartialMethods()
        {
            Console.WriteLine("エントリに対する処理を開始します");
            GeneratedEntity entity = new GeneratedEntity("FirstEntity");
            entity.FirstName = "Bob";
            entity.State = "NH";
            GeneratedEntity secondEntity = new GeneratedEntity("SecondEntity");
            entity.FirstName = "Jay";
            secondEntity.FirstName = "Steve";
            secondEntity.State = "MA";
            entity.FirstName = "Barry";
            secondEntity.State = "WA";
            secondEntity.FirstName = "Matt";
            Console.WriteLine("エントリに対する処理が終了しました");
        }

        // 出力結果：
        //エントリに対する処理を開始します
        //エンティティ FirstEntity のプロパティ(FirstName)が  から Bob に変更されました
        //エンティティ FirstEntity のプロパティ(State)が  から NH に変更されました
        //エンティティ FirstEntity のプロパティ(FirstName)が Bob から Jay に変更されました
        //エンティティ SecondEntity のプロパティ(FirstName)が  から Steve に変更されました
        //エンティティ SecondEntity のプロパティ(State)が  から MA に変更されました
        //エンティティ FirstEntity のプロパティ(FirstName)が Jay から Barry に変更されました
        //エンティティ SecondEntity のプロパティ(State)が MA から WA に変更されました
        //エンティティ SecondEntity のプロパティ(FirstName)が Steve から Matt に変更されました
        //エントリに対する処理が終了しました

        public partial class GeneratedEntity
        {
            public GeneratedEntity(string entityName)
            {
                this.EntityName = entityName;
            }

            partial void ChangingProperty(string name, string originalValue, string newValue);

            public string EntityName { get; private set; }

            private string _FirstName;
            public string FirstName
            {
                get { return _FirstName; }
                set
                {
                    ChangingProperty("FirstName", _FirstName, value);
                    _FirstName = value;
                }
            }

            private string _State;
            public string State
            {
                get { return _State; }
                set
                {
                    ChangingProperty("State", _State, value);
                    _State = value;
                }
            }
        }

        public partial class GeneratedEntity
        {
            partial void ChangingProperty(string name, string originalValue, string newValue)
            {
                Console.WriteLine("エンティティ {1} のプロパティ({0})が {2} から {3} に変更されました",
                    name, this.EntityName, originalValue, newValue);
            }
        }
        #endregion // "3.10  エンティティの生成時に行う処理を追加する"
    }
}
