﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.IO.Compression;
using System.Linq;
using System.Text;
using System.Threading;

namespace CSharpRecipes
{
    public static class FileExtensions
    {
        #region FileInfo拡張メソッド
        public static void Rename(this FileInfo originalFile, string newName)
        {
            originalFile.MoveTo(newName);
        }

        public static void DisplayTimestamps(this FileInfo fileInfo)
        {
            Console.WriteLine(fileInfo.CreationTime.ToString());
            Console.WriteLine(fileInfo.LastAccessTime.ToString());
            Console.WriteLine(fileInfo.LastWriteTime.ToString());
        }

        public static void ModifyTimestamps(this FileInfo fileInfo, DateTime dt)
        {
            fileInfo.CreationTime = dt;
            fileInfo.LastAccessTime = dt;
            fileInfo.LastWriteTime = dt;
        }

        public static void MakeFileHidden(this FileInfo fileInfo)
        {
            // ファイルの属性を変更
            fileInfo.Attributes |= FileAttributes.Hidden;
        }
        #endregion // FileInfo拡張メソッド

        #region DirectoryInfo拡張メソッド
        public static void Rename(this DirectoryInfo dirInfo, string newName)
        {
            try
            {
                // 名前を変更
                dirInfo.MoveTo(newName);
            }
            catch (IOException ioe)
            {
                // おそらく移動先のディレクトリが存在するか、空ではありません。
                Trace.WriteLine(ioe.ToString());
            }
        }

        public static void DisplayTimestamps(this DirectoryInfo dirInfo)
        {
            Console.WriteLine(dirInfo.CreationTime.ToString());
            Console.WriteLine(dirInfo.LastAccessTime.ToString());
            Console.WriteLine(dirInfo.LastWriteTime.ToString());
        }

        public static void ModifyTimestamps(this DirectoryInfo dirInfo, DateTime dt)
        {
            dirInfo.CreationTime = dt;
            dirInfo.LastAccessTime = dt;
            dirInfo.LastWriteTime = dt;
        }

        public static void MakeDirectoryHidden(this DirectoryInfo dirInfo)
        {
            // ディレクトリの属性を変更します
            dirInfo.Attributes |= FileAttributes.Hidden;
        }
        #endregion // DirectoryInfo拡張メソッド

        #region FileSystemInfo拡張メソッド
        public static string ToDisplayString(this FileSystemInfo fileSystemInfo)
        {
            string type = fileSystemInfo.GetType().ToString();
            if (fileSystemInfo is DirectoryInfo)
                type = "DIRECTORY";
            else if (fileSystemInfo is FileInfo)
                type = "FILE";
            return string.Format(Thread.CurrentThread.CurrentCulture,
                "{0}: {1}", type, fileSystemInfo.Name);
        }
        #endregion // FileSystemInfo拡張メソッド
    }

    public class FileSystemIO
    {
        #region "8.1 ファイル属性の操作"
        public static void ManipulateFileAttributes()
        {
            string path = Path.GetTempFileName();
            if (!File.Exists(path))
                File.Create(path);

            try
            {
                DateTime dt = new DateTime(2001, 2, 8);
                // Fileクラスのstaticメソッドを使用してタイムスタンプや属性を操作します
                DisplayFileTimestamps(path);
                ModifyFileTimestamps(path, dt);


                // FileInfoクラスのインスタンスを使用してタイムスタンプや属性を操作します
                FileInfo fileInfo = new FileInfo(path);
                fileInfo.DisplayTimestamps();
                fileInfo.ModifyTimestamps(dt);
                DisplayFileHiddenAttribute(path);
                fileInfo.MakeFileHidden();
                DisplayFileHiddenAttribute(path);
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
            }
        }

        public static void DisplayFileTimestamps(string path)
        {
            Console.WriteLine(File.GetCreationTime(path).ToString());
            Console.WriteLine(File.GetLastAccessTime(path).ToString());
            Console.WriteLine(File.GetLastWriteTime(path).ToString());
        }

        public static void ModifyFileTimestamps(string path, DateTime dt)
        {
            File.SetCreationTime(path, dt);
            File.SetLastAccessTime(path, dt);
            File.SetLastWriteTime(path, dt);
        }

        public static void DisplayFileHiddenAttribute(string path)
        {
            if (File.Exists(path))
            {
                FileInfo fileInfo = new FileInfo(path);

                // 現在のファイルに隠し属性が設定されているかどうかを表示します
                Console.WriteLine("隠し属性がある？ = " +
                    ((fileInfo.Attributes & FileAttributes.Hidden) == FileAttributes.Hidden));
            }
        }

        #endregion

        #region "8.2 ファイル名の変更"
        public static void RenameFile()
        {
            //File.Delete(@"C:\temp\foo.txt");
            if (!Directory.Exists(@"c:\temp"))
                Directory.CreateDirectory(@"c:\temp");
            File.Delete(@"C:\temp\bar.dat");

            if (!File.Exists(@"c:\temp\foo.txt"))
                File.Create(@"c:\temp\foo.txt");

            Rename(@"c:\temp\foo.txt", @"c:\temp\bar.dat");

            Rename(@"c:\temp\bar.dat", @"c:\temp\foo.txt");

            FileInfo originalFile = new FileInfo(@"c:\temp\foo.txt");
            originalFile.Rename(@"c:\temp\bar.dat");
        }

        public static void Rename(string originalName, string newName)
        {
            File.Move(originalName, newName);
        }

        #endregion

        #region "8.3 プラットフォームに依存しないEOL文字の出力"
        public static void PlatformIndependentEol()
        {
            // 1) すべてのテキストブロックごとにEnvironment.NewLineを使用することに注意してください。
            // ブロックの最後にはプラットフォームにおける正しい改行文字が必要です。
            string line;
            line = String.Format("FirstLine {0} SecondLine {0} ThirdLine {0}", Environment.NewLine);

            // 作業用の一時ファイルを作成します。
            string file = Path.GetTempFileName();
            FileStream stream = File.Create(file);
            byte[] bytes = Encoding.Unicode.GetBytes(line);
            stream.Write(bytes, 0, bytes.Length);
            // ファイルをクローズします
            stream.Close();

            // ファイルを削除します（この位置にブレークポイントを設定してファイルの内容を確認すると良いでしょう）
            File.Delete(file);


            // 2) テキストのwriterを設定し、特定の改行文字を指定します。

            // 新しい一時ファイルを作成します。
            file = Path.GetTempFileName();
            line = "Double spaced line";
            StreamWriter streamWriter = new StreamWriter(file);
            // 常に改行2つで書き出すようにします。
            streamWriter.NewLine = Environment.NewLine + Environment.NewLine;
            // このストリームにおけるWriteLineは、新たに指定された改行コードを
            // 自動的に使用するようになります（この場合は改行2つ）
            streamWriter.WriteLine(line);
            streamWriter.WriteLine(line);
            streamWriter.WriteLine(line);
            // ファイルをクローズします
            streamWriter.Close();

            // ファイルを削除します（この位置にブレークポイントを設定してファイルの内容を確認すると良いでしょう）
            File.Delete(file);

            // 3) 通常のWriteLineメソッドはすべてデフォルトでEnvironment.NewLineを使用します。

            line = "Default line";
            Console.WriteLine(line);

        }
        #endregion

        #region "8.4 ディレクトリ属性の操作"
        public static void ManipulateDirectoryAttributes()
        {
            string path = Path.GetTempPath() + @"\MyTemp";
            Directory.CreateDirectory(path);
            try
            {
                DateTime dt = new DateTime(2003, 5, 10);
                // Directoryクラスのstaticメソッドを使用してディレクトリのタイムスタンプや属性を操作します
                DisplayDirectoryTimestamps(path);
                ModifyDirectoryTimestamps(path, dt);


                // DirectoryInfoクラスのインスタンスを使用してディレクトリのタイムスタンプや属性を操作します
                DirectoryInfo dirInfo = new DirectoryInfo(path);
                dirInfo.DisplayTimestamps();
                dirInfo.ModifyTimestamps(dt);
                DisplayDirectoryHiddenAttribute(path);
                dirInfo.MakeDirectoryHidden();
                DisplayDirectoryHiddenAttribute(path);
            }
            catch (Exception e)
            {
                Console.WriteLine("{0}", e.ToString());
            }
            Directory.Delete(path, true);
        }

        public static void DisplayDirectoryTimestamps(string path)
        {
            Console.WriteLine(Directory.GetCreationTime(path).ToString());
            Console.WriteLine(Directory.GetLastAccessTime(path).ToString());
            Console.WriteLine(Directory.GetLastWriteTime(path).ToString());
        }

        public static void ModifyDirectoryTimestamps(string path, DateTime dt)
        {
            Directory.SetCreationTime(path, dt);
            Directory.SetLastAccessTime(path, dt);
            Directory.SetLastWriteTime(path, dt);
        }

        public static void DisplayDirectoryHiddenAttribute(string path)
        {
            DirectoryInfo dirInfo = new DirectoryInfo(path);
            // ディレクトリに隠し属性があるかどうかを表示します
            Console.WriteLine("ディレクトリに隠し属性がある？ = " +
                ((dirInfo.Attributes & FileAttributes.Hidden) == FileAttributes.Hidden));
        }

        #endregion

        #region "8.5 ディレクトリ名の変更"
        public static void RenameDirectory()
        {
            string path = Path.GetTempPath() + "MyTemp";
            string newpath = path + "1";
            Directory.CreateDirectory(path);

            RenameDirectory(path, newpath);
            DirectoryInfo dirInfo = new DirectoryInfo(newpath);
            dirInfo.Rename(path);
        }

        public static void RenameDirectory(string originalName, string newName)
        {
            try
            {
                // 名前を変更します
                Directory.Move(originalName, newName);
            }
            catch (IOException ioe)
            {
                // おそらく移動先のディレクトリが存在するか、空ではありません。
                Console.WriteLine(ioe.ToString());
            }
        }

        #endregion

        #region "8.6 ワイルドカードを使用したディレクトリまたはファイル検索"
        public static void SearchDirFileWildcards()
        {
            try
            {
                // 各パスの初期値を設定します
                string tempPath = Path.GetTempPath();
                string tempDir = tempPath + @"\MyTempDir";
                string tempDir2 = tempDir + @"\MyNestedTempDir";
                string tempDir3 = tempDir + @"\MyNestedTempDirPattern";
                string tempFile = tempDir + @"\MyTempFile.PDB";
                string tempFile2 = tempDir2 + @"\MyNestedTempFile.PDB";
                string tempFile3 = tempDir + @"\MyTempFile.TXT";

                // 一時ディレクトリおよびファイルを作成します
                Directory.CreateDirectory(tempDir);
                Directory.CreateDirectory(tempDir2);
                Directory.CreateDirectory(tempDir3);
                FileStream stream = File.Create(tempFile);
                FileStream stream2 = File.Create(tempFile2);
                FileStream stream3 = File.Create(tempFile3);
                // ファイルを使用する前にクローズします
                stream.Close();
                stream2.Close();
                stream3.Close();

                // tempディレクトリ以下のすべてのファイルまたはディレクトリを表示します
                DisplayFilesForADirectory(tempDir);

                // パターンに一致するファイルまたはディレクトリのみを表示します
                string pattern = "*.PDB";
                DisplayFilesWithPattern(tempDir, pattern);

                // tempディレクトリ以下のすべてのディレクトリを表示します
                DisplayDirectories(tempDir);

                // パターンに一致するディレクトリのみを表示します
                pattern = "*Pattern*";
                DisplayDirectoriesWithPattern(tempDir, pattern);

                // tempディレクトリ以下のすべてのファイルを表示します
                DisplayFiles1(tempDir);

                // パターンに一致するファイルのみを表示します
                pattern = "*.PDB";
                DisplayFiles3(tempDir, pattern);

                // 以下ではパスの文字列を返すのではなく、パスが表す実際のオブジェクトを使用します

                // すべてのファイルまたはディレクトリを取得します
                DisplayDirectoryContents(tempDir);

                // パターンに一致するすべてのファイルまたはディレクトリを取得します
                pattern = "*.PDB";
                DisplayDirectoryContentsWithPattern(tempDir, pattern);

                // すべてのDirectoryInfoを取得します
                DisplayDirectoriesFromInfo(tempDir);

                // パターンに一致するDirectoryInfoのみを取得します
                pattern = "*Pattern*";
                DisplayDirectoriesWithPattern(tempDir, pattern);

                // すべてのFileInfoを取得します
                DisplayFiles2(tempDir);

                // パターンに一致するFileInfoのみを取得します
                pattern = "*.PDB";
                DisplayFiles4(tempDir, pattern);

                // テスト時に使用したファイルまたはディレクトリを削除します
                File.Delete(tempFile3);
                File.Delete(tempFile2);
                File.Delete(tempFile);
                Directory.Delete(tempDir3);
                Directory.Delete(tempDir2);
                Directory.Delete(tempDir);
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
            }
        }

        public static void DisplayFilesForADirectory(string path)
        {
            string[] items = Directory.GetFileSystemEntries(path);
            foreach (string item in items)
            {
                Console.WriteLine(item);
            }
        }

        public static void DisplayDirectories(string path)
        {
            string[] items = Directory.GetDirectories(path);
            foreach (string item in items)
            {
                Console.WriteLine(item);
            }
        }

        public static void DisplayFiles1(string path)
        {
            string[] items = Directory.GetFiles(path);
            foreach (string item in items)
            {
                Console.WriteLine(item);
            }
        }

        public static void DisplayDirectoryContents(string path)
        {
            DirectoryInfo mainDir = new DirectoryInfo(path);
            IEnumerable<string> fileSystemDisplayInfos =
                from fsi in mainDir.GetFileSystemInfos()
                where fsi is FileSystemInfo || fsi is DirectoryInfo
                select fsi.ToDisplayString();

            foreach (string s in fileSystemDisplayInfos)
            {
                Console.WriteLine(s);
            }
        }

        public static void DisplayDirectoriesFromInfo(string path)
        {
            DirectoryInfo mainDir = new DirectoryInfo(path);
            DirectoryInfo[] items = mainDir.GetDirectories();
            foreach (DirectoryInfo item in items)
            {
                Console.WriteLine("DIRECTORY: " + item.Name);
            }
        }

        public static void DisplayFiles2(string path)
        {
            DirectoryInfo mainDir = new DirectoryInfo(path);
            FileInfo[] items = mainDir.GetFiles();
            foreach (FileInfo item in items)
            {
                Console.WriteLine("FILE: " + item.Name);
            }
        }

        // テストではパターン"*.PDB"が指定されます
        public static void DisplayFilesWithPattern(string path, string pattern)
        {
            string[] items = Directory.GetFileSystemEntries(path, pattern);
            foreach (string item in items)
            {
                Console.WriteLine(item);
            }
        }

        // テストではパターン"*.PDB"が指定されます
        public static void DisplayDirectoriesWithPattern(string path, string pattern)
        {
            string[] items = Directory.GetDirectories(path, pattern);
            foreach (string item in items)
            {
                Console.WriteLine(item);
            }
        }

        // テストではパターン"*.PDB"が指定されます
        public static void DisplayFiles3(string path, string pattern)
        {
            string[] items = Directory.GetFiles(path, pattern);
            foreach (string item in items)
            {
                Console.WriteLine(item);
            }
        }

        // テストではパターン"*.PDB"が指定されます
        public static void DisplayDirectoryContentsWithPattern(string path, string pattern)
        {
            DirectoryInfo mainDir = new DirectoryInfo(path);
            IEnumerable<string> fileSystemDisplayInfos =
                from fsi in mainDir.GetFileSystemInfos(pattern)
                where fsi is FileSystemInfo || fsi is DirectoryInfo
                select fsi.ToDisplayString();

            foreach (string s in fileSystemDisplayInfos)
            {
                Console.WriteLine(s);
            }
        }

        // テストではパターン"*.PDB"が指定されます
        public static void DisplayDirectoriesWithPatternFromInfo(string path, string pattern)
        {
            DirectoryInfo mainDir = new DirectoryInfo(path);
            DirectoryInfo[] items = mainDir.GetDirectories(pattern);
            foreach (DirectoryInfo item in items)
            {
                Console.WriteLine("DIRECTORY: " + item.Name);
            }
        }

        // テストではパターン"*.PDB"が指定されます
        public static void DisplayFiles4(string path, string pattern)
        {
            DirectoryInfo mainDir = new DirectoryInfo(path);
            FileInfo[] items = mainDir.GetFiles(pattern);
            foreach (FileInfo item in items)
            {
                Console.WriteLine("FILE: " + item.Name);
            }
        }

        #endregion

        #region "8.7 ディレクトリ構造の取得"
        public static void ObtainDirTree()
        {
            try
            {
                // 検索対象のディレクトリを設定します
                string tempPath = Path.GetTempPath();
                string tempDir = tempPath + @"\MyTempDir";
                string tempDir2 = tempDir + @"\Chapter 1 - The Beginning";
                string tempDir3 = tempDir2 + @"\Chapter 1a - The Rest Of The Beginning";
                string tempDir4 = tempDir2 + @"\IHaveAPDBFile";
                string tempFile = tempDir + @"\MyTempFile.PDB";
                string tempFile2 = tempDir2 + @"\MyNestedTempFile.PDB";
                string tempFile3 = tempDir3 + @"\Chapter 1 - MyDeepNestedTempFile.PDB";
                string tempFile4 = tempDir4 + @"\PDBFile.PDB";
                // ディレクトリ構造は以下のようになります
                // TEMP\
                //		MyTempDir\
                //				Chapter 1 - The Beginning\
                //								Chapter 1a - The Rest Of The Beginning\
                //								IHaveAPDBFile
                //

                // 一時ディレクトリおよびファイルを作成します
                Directory.CreateDirectory(tempDir);
                Directory.CreateDirectory(tempDir2);
                Directory.CreateDirectory(tempDir3);
                Directory.CreateDirectory(tempDir4);
                FileStream stream = File.Create(tempFile);
                FileStream stream2 = File.Create(tempFile2);
                FileStream stream3 = File.Create(tempFile3);
                FileStream stream4 = File.Create(tempFile4);
                // ファイルを使用する前にクローズします
                stream.Close();
                stream2.Close();
                stream3.Close();
                stream4.Close();

                //tempDir = Environment.GetFolderPath(Environment.SpecialFolder.ProgramFiles);
                // 再帰呼び出しによりすべてのファイルを一覧表示します
                Stopwatch watch1 = Stopwatch.StartNew();
                DisplayAllFilesAndDirectories(tempDir);
                watch1.Stop();
                Console.WriteLine("*************************");

                // 再帰呼び出しなしですべてのファイルを一覧表示します
                Stopwatch watch2 = Stopwatch.StartNew();
                DisplayAllFilesAndDirectoriesWithoutRecursion(tempDir);
                watch2.Stop();
                Console.WriteLine("*************************");
                Console.WriteLine("Recursive method time elapsed {0}", watch1.Elapsed.ToString());
                Console.WriteLine("Non-Recursive method time elapsed {0}", watch2.Elapsed.ToString());

                //// 拡張子がPDBのファイルに限定して一覧表示します
                DisplayAllFilesWithExtensionWithoutRecursion(tempDir, ".PDB");

                // テスト時に使用したファイルまたはディレクトリを削除します
                File.Delete(tempFile4);
                File.Delete(tempFile3);
                File.Delete(tempFile2);
                File.Delete(tempFile);
                Directory.Delete(tempDir4);
                Directory.Delete(tempDir3);
                Directory.Delete(tempDir2);
                Directory.Delete(tempDir);
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
            }
        }



        public static void DisplayAllFilesAndDirectories(string dir)
        {
            var strings = from fileSystemInfo in GetAllFilesAndDirectories(dir)
                          select fileSystemInfo.ToDisplayString();

            foreach (string s in strings)
                Console.WriteLine(s);
        }

        public static void DisplayAllFilesWithExtension(string dir, string extension)
        {
            var strings = from fileSystemInfo in GetAllFilesAndDirectories(dir)
                          where fileSystemInfo is FileInfo &&
                                fileSystemInfo.FullName.Contains("Chapter 1") &&
                                (string.Compare(fileSystemInfo.Extension, extension,
                                                StringComparison.OrdinalIgnoreCase) == 0)
                          select fileSystemInfo.ToDisplayString();

            foreach (string s in strings)
                Console.WriteLine(s);
        }


        public static IEnumerable<FileSystemInfo> GetAllFilesAndDirectories(string dir)
        {
            DirectoryInfo dirInfo = new DirectoryInfo(dir);
            FileSystemInfo[] fileSystemInfos = dirInfo.GetFileSystemInfos();
            foreach (FileSystemInfo fileSystemInfo in fileSystemInfos)
            {
                yield return fileSystemInfo;
                if (fileSystemInfo is DirectoryInfo)
                {
                    foreach (FileSystemInfo fsi in GetAllFilesAndDirectories(fileSystemInfo.FullName))
                        yield return fsi;
                }
            }
        }

        public static void DisplayAllFilesAndDirectoriesWithoutRecursion(string dir)
        {
            var strings = from fileSystemInfo in GetAllFilesAndDirectoriesWithoutRecursion(dir)
                          select fileSystemInfo.ToDisplayString();

            foreach (string s in strings)
                Console.WriteLine(s);
        }

        public static void DisplayAllFilesWithExtensionWithoutRecursion(string dir, string extension)
        {
            var strings = from fileSystemInfo in GetAllFilesAndDirectoriesWithoutRecursion(dir)
                          where fileSystemInfo is FileInfo &&
                                fileSystemInfo.FullName.Contains("Chapter 1") &&
                                (string.Compare(fileSystemInfo.Extension, extension,
                                                StringComparison.OrdinalIgnoreCase) == 0)
                          select fileSystemInfo.ToDisplayString();

            foreach (string s in strings)
                Console.WriteLine(s);
        }

        public static IEnumerable<FileSystemInfo> GetAllFilesAndDirectoriesWithoutRecursion(string dir)
        {
            DirectoryInfo dirInfo = new DirectoryInfo(dir);
            Stack<FileSystemInfo> stack = new Stack<FileSystemInfo>();

            stack.Push(dirInfo);
            while (dirInfo != null || stack.Count > 0)
            {
                FileSystemInfo fileSystemInfo = stack.Pop();
                DirectoryInfo subDirectoryInfo = fileSystemInfo as DirectoryInfo;
                if (subDirectoryInfo != null)
                {
                    yield return subDirectoryInfo;
                    foreach (FileSystemInfo fsi in subDirectoryInfo.GetFileSystemInfos())
                        stack.Push(fsi);
                    dirInfo = subDirectoryInfo;
                }
                else
                {
                    yield return fileSystemInfo;
                    dirInfo = null;
                }
            }
        }

        #endregion

        #region "8.8 パスの解析"
        public static void ParsePath()
        {
            ParsePath(@"c:\test\tempfile.txt");
        }

        public static void ParsePath(string path)
        {
            string root = Path.GetPathRoot(path);
            string dirName = Path.GetDirectoryName(path);
            string fullFileName = Path.GetFileName(path);
            string fileExt = Path.GetExtension(path);
            string fileNameWithoutExt = Path.GetFileNameWithoutExtension(path);
            StringBuilder format = new StringBuilder();
            format.Append("パス「{0}」は以下のように分解されます:" + Environment.NewLine +
                          "\tルート: {1}" + Environment.NewLine + "\t");
            format.Append("ディレクトリ名: {2}" + Environment.NewLine +
                           "\tファイル名: {3}" + Environment.NewLine + "\t");
            format.Append("ファイル拡張子: {4}" + Environment.NewLine +
                          "\t拡張子なしのファイル名: {5}" + Environment.NewLine + "");
            Console.WriteLine(format.ToString(), path, root, dirName,
                              fullFileName, fileExt, fileNameWithoutExt);
        }

        #endregion

        #region "8.9 環境変数におけるパスの解析"
        public static void ParsePathsEnvironmentVariables()
        {
            ParsePathEnvironmentVariable();
        }

        public static void ParsePathEnvironmentVariable()
        {
            string originalPathEnv = Environment.GetEnvironmentVariable("PATH");
            string[] paths = originalPathEnv.Split(Path.PathSeparator);
            foreach (string s in paths)
            {
                string pathEnv = Environment.ExpandEnvironmentVariables(s);
                if (!string.IsNullOrEmpty(pathEnv))
                    Console.WriteLine("Individual Path = " + pathEnv);
                Console.WriteLine();
            }
        }

        #endregion

        #region "8.10 コンソールアプリケーションの起動および利用方法"
        public static void LaunchInteractConsoleUtilities()
        {
            RunProcessToReadStdIn();
        }

        public static void RunProcessToReadStdIn()
        {
            Process application = new Process();
            // コマンドシェルを起動します。
            application.StartInfo.FileName = @"cmd.exe";

            // cmd.exe用の拡張機能を有効にします。
            application.StartInfo.Arguments = "/E:ON";

            application.StartInfo.RedirectStandardInput = true;

            application.StartInfo.UseShellExecute = false;

            // プロセスを開始します
            application.Start();

            // 標準出力を保持します
            StreamWriter input = application.StandardInput;

            // 時刻表示用のコマンドを実行します
            input.WriteLine("TIME /T");

            // 実行中のコマンドを終了します
            input.WriteLine("exit");
        }

        #endregion

        #region "8.11 ファイルの排他処理"
        public static void LockSubsectionsOfAFile()
        {
            string path = Path.GetTempFileName();
            CreateLockedFile(path);
            File.Delete(path);
        }

        public static void CreateLockedFile(string fileName)
        {
            using (FileStream fileStream = new FileStream(fileName,
                      FileMode.Create,
                      FileAccess.ReadWrite,
                      FileShare.ReadWrite))
            {
                using (StreamWriter streamWriter = new StreamWriter(fileStream))
                {
                    streamWriter.WriteLine("一行目");
                    streamWriter.WriteLine("二行目");
                    streamWriter.Flush();

                    try
                    {
                        // ファイルの全体をロックします。
                        fileStream.Lock(0, fileStream.Length);

                        // 時間のかかる処理をここで行います
                        Thread.Sleep(1000);
                    }
                    finally
                    {
                        // 必ずファイルのロックが解除されるようにします。
                        // ファイルの一部をロックしたままプロセスが終了したり、
                        // 未解決のロックを残したままでファイルを閉じたりした場合の動作は
                        // 未定義であり、Microsoftはその動作を保証していません。
                        fileStream.Unlock(0, fileStream.Length);
                    }

                    streamWriter.WriteLine("三行目");
                }
            }
        }

        #endregion

        #region "8.12 ファイルシステム内アクションの待機"
        public static void WaitFileSystemAction()
        {
            string tempPath = Path.GetTempPath();
            string fileName = "temp.zip";
            File.Delete(tempPath + fileName);

            WaitForZipCreation(tempPath, fileName);
        }

        public static void WaitForZipCreation(string path, string fileName)
        {
            FileSystemWatcher fsw = null;
            try
            {
                fsw = new FileSystemWatcher();
                string[] data = new string[] { path, fileName };
                fsw.Path = path;
                fsw.Filter = fileName;
                fsw.NotifyFilter = NotifyFilters.LastAccess | NotifyFilters.LastWrite
                    | NotifyFilters.FileName | NotifyFilters.DirectoryName;

                // 検出対象のファイルを生成するコードを実行します。
                // 普通は異なるソースによってい対象ファイルが生成されるため、
                // 同一コード内でこの処理を行う必要はありません。
                if (ThreadPool.QueueUserWorkItem(new WaitCallback(PauseAndCreateFile),
                    data))
                {
                    Thread.Sleep(2000);
                    // 変更を待機
                    WaitForChangedResult result =
                        fsw.WaitForChanged(WatcherChangeTypes.Created);
                    Console.WriteLine("{0} created at {1}.", result.Name, path);
                }
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
            }

            // 後処理を行います
            File.Delete(fileName);

            if (fsw != null)
                fsw.Dispose();
        }

        static void PauseAndCreateFile(Object stateInfo)
        {
            try
            {
                string[] data = (string[])stateInfo;
                // しばらく待ちます...
                Thread.Sleep(1000);
                // 一時ディレクトリにファイルを作成します
                string path = data[0];
                string file = path + data[1];
                Console.WriteLine("PauseAndCreateFileにおいて{0}を作成中...", file);
                FileStream fileStream = File.Create(file);
                fileStream.Close();
            }
            catch (Exception e)
            {
                Console.WriteLine(e.ToString());
            }
        }
        #endregion

        #region "8.13 実行可能モジュールのバージョン情報の比較"
        public static void CompareVersionInfo()
        {
            // バージョンの比較
            string file1 = Path.GetTempFileName();
            string file2 = Path.GetTempFileName();
            FileSystemIO.FileComparison result =
                FileSystemIO.CompareFileVersions(file1, file2);
            // 後処理
            File.Delete(file1);
            File.Delete(file2);
        }

        public enum FileComparison
        {
            Error = 0,
            Newer = 1,
            Older = 2,
            Same = 3
        }


        private static FileComparison ComparePart(int p1, int p2)
        {
            return p1 > p2 ? FileComparison.Newer : (p1 < p2 ? FileComparison.Older : FileComparison.Same);
        }

        public static FileComparison CompareFileVersions(string file1, string file2)
        {
            FileComparison retValue = FileComparison.Error;

            if (!File.Exists(file1) || !File.Exists(file2))
            {
                // 比較対象のファイルが見つかりませんでした
                return retValue;
            }

            // バージョン情報を取得します
            FileVersionInfo file1Version = FileVersionInfo.GetVersionInfo(file1);
            FileVersionInfo file2Version = FileVersionInfo.GetVersionInfo(file2);

            retValue = ComparePart(file1Version.FileMajorPart, file2Version.FileMajorPart);
            if (retValue != FileComparison.Same)
            {
                retValue = ComparePart(file1Version.FileMinorPart, file2Version.FileMinorPart);
                if (retValue != FileComparison.Same)
                {
                    retValue = ComparePart(file1Version.FileBuildPart, file2Version.FileBuildPart);
                    if (retValue != FileComparison.Same)
                        retValue = ComparePart(file1Version.FilePrivatePart, file2Version.FilePrivatePart);
                }
            }
            return retValue;
        }

        #endregion

        #region "8.14 システム内のドライブへの問い合わせ"
        public static void TestAllDriveInfo()
        {
            DisplayOneDrivesInfo();
            DisplayAllDriveInfo();
        }

        public static void DisplayOneDrivesInfo()
        {
            DriveInfo drive = new DriveInfo("D");
            if (drive.IsReady)
                Console.WriteLine("D:\\ドライブ上の空き容量：" + drive.AvailableFreeSpace);
            else
                Console.WriteLine("ドライブ D:\\ は使用できません。");

            // ドライブが使用できない場合、以下の例外が発生します：
            //   System.IO.IOException: デバイスの準備ができていません。
            Console.WriteLine();
        }

        public static void DisplayAllDriveInfo()
        {
            foreach (DriveInfo drive in DriveInfo.GetDrives())
            {
                if (drive.IsReady)
                {
                    Console.WriteLine("ドライブ " + drive.Name + " は使用可能です。");
                    Console.WriteLine("AvailableFreeSpace: " + drive.AvailableFreeSpace);
                    Console.WriteLine("DriveFormat: " + drive.DriveFormat);
                    Console.WriteLine("DriveType: " + drive.DriveType);
                    Console.WriteLine("Name: " + drive.Name);
                    Console.WriteLine("RootDirectory.FullName: " + drive.RootDirectory.FullName);
                    Console.WriteLine("TotalFreeSpace: " + drive.TotalFreeSpace);
                    Console.WriteLine("TotalSize: " + drive.TotalSize);
                    Console.WriteLine("VolumeLabel: " + drive.VolumeLabel);
                }
                else
                {
                    Console.WriteLine("ドライブ " + drive.Name + " は使用できません。");
                }

                Console.WriteLine();
            }
        }
        #endregion

        #region "8.15 ファイルの圧縮と解凍"
        public static void TestCompressNewFile()
        {
            byte[] data = new byte[10000000];
            for (int i = 0; i < 10000000; i++)
                data[i] = (byte)i;

            FileStream fs =
                new FileStream(@"C:\NewNormalFile.txt",
                    FileMode.OpenOrCreate, FileAccess.ReadWrite, FileShare.None);
            using (fs)
            {
                fs.Write(data, 0, data.Length);
            }

            CompressFile(@"C:\NewNormalFile.txt", @"C:\NewCompressedFile.txt",
                CompressionType.Deflate);

            DecompressFile(@"C:\NewCompressedFile.txt", @"C:\NewDecompressedFile.txt",
                CompressionType.Deflate);

            CompressFile(@"C:\NewNormalFile.txt", @"C:\NewGZCompressedFile.txt",
                CompressionType.GZip);

            DecompressFile(@"C:\NewGZCompressedFile.txt", @"C:\NewGZDecompressedFile.txt",
                CompressionType.GZip);

            //Normal file size == 10,000,000 bytes
            //GZipped file size == 155,204
            //Deflated file size == 155,168
            // 36 bytes are related to the GZip CRC
        }

        /// <summary>
        /// 指定したファイルを新しいファイルに圧縮します。
        /// メモリ消費量を抑えるために、1MB単位で処理をします。
        /// </summary>
        /// <param name="sourceFile">未圧縮のファイル</param>
        /// <param name="destinationFile">圧縮されたファイル</param>
        /// <param name="compressionType">使用する圧縮タイプ</param>
        public static void CompressFile(string sourceFile,
                                        string destinationFile,
                                        CompressionType compressionType)
        {
            if (sourceFile != null)
            {
                FileStream streamSource = null;
                FileStream streamDestination = null;
                Stream streamCompressed = null;

                using (streamSource = File.OpenRead(sourceFile))
                {
                    using (streamDestination = File.OpenWrite(destinationFile))
                    {
                        // 1MBのデータのかたまり（チャンク）を読み込んで圧縮します。
                        long fileLength = streamSource.Length;

                        // fileLengthの大きさを書き出します
                        byte[] size = BitConverter.GetBytes(fileLength);
                        streamDestination.Write(size, 0, size.Length);

                        long chunkSize = 1048576; // 1MB
                        while (fileLength > 0)
                        {
                            // チャンクを読み込みます
                            byte[] data = new byte[chunkSize];
                            streamSource.Read(data, 0, data.Length);

                            // チャンクを圧縮します
                            MemoryStream compressedDataStream =
                                new MemoryStream();

                            if (compressionType == CompressionType.Deflate)
                                streamCompressed =
                                    new DeflateStream(compressedDataStream,
                                        CompressionMode.Compress);
                            else
                                streamCompressed =
                                    new GZipStream(compressedDataStream,
                                        CompressionMode.Compress);

                            using (streamCompressed)
                            {
                                // 圧縮ファイル用ストリームにチャンクを書き出します
                                streamCompressed.Write(data, 0, data.Length);
                            }
                            // 圧縮されたチャンクのバイト数を取得します
                            byte[] compressedData =
                                compressedDataStream.GetBuffer();

                            // チャンクのサイズを書き出します
                            size = BitConverter.GetBytes(chunkSize);
                            streamDestination.Write(size, 0, size.Length);

                            // 圧縮後のサイズを書き出します
                            size = BitConverter.GetBytes(compressedData.Length);
                            streamDestination.Write(size, 0, size.Length);

                            // 圧縮されたチャンクを書き出します
                            streamDestination.Write(compressedData, 0,
                                compressedData.Length);

                            // ファイルサイズからチャンクのサイズを差し引きます
                            fileLength -= chunkSize;

                            // 残りのファイルサイズがチャンクのサイズより小さくなったら、
                            // 残りファイルの大きさをチャンクの大きさとします
                            if (fileLength < chunkSize)
                                chunkSize = fileLength;
                        }
                    }
                }
            }
        }

        /// <summary>
        /// このメソッドは、CompressFileメソッドで作成した
        /// チャンクに分割された圧縮ファイルを解凍します。
        /// </summary>
        /// <param name="sourceFile">圧縮されたファイル</param>
        /// <param name="destinationFile">出力ファイル</param>
        /// <param name="compressionType">使用された圧縮タイプ</param>
        public static void DecompressFile(string sourceFile,
                                        string destinationFile,
                                        CompressionType compressionType)
        {
            FileStream streamSource = null;
            FileStream streamDestination = null;
            Stream streamUncompressed = null;

            using (streamSource = File.OpenRead(sourceFile))
            {
                using (streamDestination = File.OpenWrite(destinationFile))
                {
                    // ファイルのサイズを読み込みます
                    // チャンクのサイズを読み込みます
                    byte[] size = new byte[sizeof(long)];
                    streamSource.Read(size, 0, size.Length);
                    // サイズを数字に変換します
                    long fileLength = BitConverter.ToInt64(size, 0);
                    long chunkSize = 0;
                    int storedSize = 0;
                    long workingSet = Process.GetCurrentProcess().WorkingSet64;
                    while (fileLength > 0)
                    {
                        // チャンクのサイズを読み込みます
                        size = new byte[sizeof(long)];
                        streamSource.Read(size, 0, size.Length);
                        // サイズを数字に変換します
                        chunkSize = BitConverter.ToInt64(size, 0);
                        if (chunkSize > fileLength ||
                            chunkSize > workingSet)
                            throw new InvalidDataException();

                        // 圧縮されたサイズを読み込みます
                        size = new byte[sizeof(int)];
                        streamSource.Read(size, 0, size.Length);
                        // サイズを数字に変換します
                        storedSize = BitConverter.ToInt32(size, 0);
                        if (storedSize > fileLength ||
                            storedSize > workingSet)
                            throw new InvalidDataException();

                        if (storedSize > chunkSize)
                            throw new InvalidDataException();

                        byte[] uncompressedData = new byte[chunkSize];
                        byte[] compressedData = new byte[storedSize];
                        streamSource.Read(compressedData, 0,
                            compressedData.Length);

                        // チャンクを解凍します
                        MemoryStream uncompressedDataStream =
                            new MemoryStream(compressedData);

                        if (compressionType == CompressionType.Deflate)
                            streamUncompressed =
                                new DeflateStream(uncompressedDataStream,
                                    CompressionMode.Decompress);
                        else
                            streamUncompressed =
                                new GZipStream(uncompressedDataStream,
                                    CompressionMode.Decompress);

                        using (streamUncompressed)
                        {
                            // 圧縮ファイル用のストリームからチャンクを読み込みます
                            streamUncompressed.Read(uncompressedData, 0,
                                uncompressedData.Length);
                        }

                        // 解凍されたチャンクを書き出します
                        streamDestination.Write(uncompressedData, 0,
                            uncompressedData.Length);

                        // ファイルサイズからチャンクのサイズを差し引きます
                        fileLength -= chunkSize;

                        // 残りのファイルサイズがチャンクのサイズより小さくなったら、
                        // 残りファイルの大きさをチャンクの大きさとします。
                        if (fileLength < chunkSize)
                            chunkSize = fileLength;
                    }
                }
            }
        }

        public enum CompressionType
        {
            Deflate,
            GZip
        }
        #endregion

    }
}
