﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Xml.Linq;
using System.Reflection.Emit;

[assembly: ObfuscateAssembly(true, StripAfterObfuscation = true)]

namespace CSharpRecipes
{
    public static class ReflectionExt
    {
        #region 9.6 拡張メソッド
        public static string GetBaseTypeDisplay(this Type type)
        {
            IEnumerable<string> baseTypes =
                (from t in type.GetBaseTypes()
                 select t.Name).Reverse();
            StringBuilder builder = new StringBuilder();
            foreach (string typeName in baseTypes)
            {
                if (builder.Length == 0)
                    builder.Append(typeName);
                else
                    builder.AppendFormat("<-{0}", typeName);
            }
            return builder.ToString();
        }

        private static IEnumerable<Type> GetBaseTypes(this Type type)
        {
            Type current = type;
            while (current != null)
            {
                yield return current;
                current = current.BaseType;
            }
        }
        #endregion

        #region 9.7 拡張メソッド
        public static IEnumerable<Type> ListSubClassesForType(this Assembly asm,
                                                              Type baseClassType)
        {
            return from type in asm.GetTypes()
                   where type.IsSubclassOf(baseClassType)
                   select type;
        }
        #endregion
    }

    public class Reflection
    {
        #region "9.1 参照先のアセンブリ一覧を取得する"
        public static void ListImportedAssemblies()
        {
            string file = GetProcessPath();

            //ProcessModuleCollection pmc = curProc.Modules;
            //foreach (ProcessModule pm in pmc)
            //{
            //    Console.WriteLine("ModuleName: " + pm.ModuleName);
            //    Console.WriteLine("FileName: " + pm.FileName);
            //    Console.WriteLine("ModuleName: " + pm.FileVersionInfo.FileName);
            //}

            StringCollection assemblies = new StringCollection();
            Reflection.BuildDependentAssemblyList(file, assemblies);
            Console.WriteLine("アセンブリ {0} の依存ツリーは以下の通りです：\r\n", file);
            foreach (string name in assemblies)
            {
                Console.WriteLine("\t{0}\r\n", name);
            }
        }

        public static void BuildDependentAssemblyList(string path,
            StringCollection assemblies)
        {
            // 元のアセンブリが参照するアセンブリの一覧を管理します
            if (assemblies == null)
                assemblies = new StringCollection();

            // 既にリストに含まれている？
            if (assemblies.Contains(path) == true)
                return;

            try
            {
                Assembly asm = null;

                // 文字列中から一般的な区切り文字を探して
                // ファイルかディレクトリかを判断します
                if ((path.IndexOf(@"\", 0, path.Length, StringComparison.Ordinal) != -1) ||
                    (path.IndexOf("/", 0, path.Length, StringComparison.Ordinal) != -1))
                {
                    // 指定されたパスからアセンブリを読み込みます
                    asm = Assembly.LoadFrom(path);
                }
                else
                {
                    // アセンブリ名として読み込みを試みます
                    asm = Assembly.Load(path);
                }

                // アセンブリをリストに追加します
                if (asm != null)
                {
                    assemblies.Add(path);
                }

                // 参照先のアセンブリを取得します
                AssemblyName[] imports = asm.GetReferencedAssemblies();

                // 列挙します
                foreach (AssemblyName asmName in imports)
                {
                    // このメソッドを再帰呼び出しすることによって
                    // 参照先のアセンブリがさらに別のアセンブリを参照しているかどうかを調べます
                    BuildDependentAssemblyList(asmName.FullName, assemblies);
                }
            }
            catch (FileLoadException fle)
            {
                // とりあえず処理を先に進めます...
                Console.WriteLine(fle);
            }
        }

        #endregion

        #region "9.2 外部公開された型の一覧を取得する"
        public static void ListExportedTypes()
        {
            string file = GetProcessPath();
            ListExportedTypes(file);
        }

        public static void ListExportedTypes(string path)
        {
            // アセンブリをロードします
            Assembly asm = Assembly.LoadFrom(path);
            Console.WriteLine("アセンブリ {0}：", path);
            // 公開されている型の一覧を取得します
            Type[] types = asm.GetExportedTypes();
            foreach (Type t in types)
            {
                Console.WriteLine("\t公開されている型：{0}", t.FullName);
            }
        }

        #endregion

        #region "9.3 オーバーライドされたメソッドの検出"
        public abstract class BaseOverrides
        {
            public abstract void Foo(string str, int i);

            public abstract void Foo(long l, double d, byte[] bytes);
        }

        public class DerivedOverrides : BaseOverrides
        {
            public override void Foo(string str, int i)
            {
            }

            public override void Foo(long l, double d, byte[] bytes)
            {
            }
        }

        private static string GetProcessPath()
        {
            // デバッガ経由で実行されている場合、元々の実行ファイル名となるように調整します
            string processName = Process.GetCurrentProcess().MainModule.FileName;
            int index = processName.IndexOf("vshost", StringComparison.Ordinal);
            if (index != -1)
            {
                string first = processName.Substring(0, index);
                int numChars = processName.Length - (index + 7);
                string second = processName.Substring(index + 7, numChars);

                processName = first + second;
            }
            return processName;
        }

        public static void FindOverriddenMethods()
        {
            string path = GetProcessPath();

            // 簡易版のオーバーロードを実行します
            FindMethodOverrides(path, "CSharpRecipes.Reflection+DerivedOverrides");

            // シグネチャを指定して実行します
            FindMethodOverrides(path,
                "CSharpRecipes.Reflection+DerivedOverrides",
                "Foo",
                new Type[3] { typeof(long), typeof(double), typeof(byte[]) });
        }

        public static void FindMethodOverrides(string asmPath, string typeName)
        {
            Assembly asm = Assembly.LoadFrom(asmPath);
            Type asmType = asm.GetType(typeName);

            Console.WriteLine("---[" + asmType.FullName + "]---");

            // 現在の型からメソッドを取得します
            MethodInfo[] methods = asmType.GetMethods(BindingFlags.Instance |
                BindingFlags.NonPublic | BindingFlags.Public |
                BindingFlags.Static | BindingFlags.DeclaredOnly);

            var mis = from ms in methods
                      where ms != ms.GetBaseDefinition()
                      select ms.GetBaseDefinition();

            foreach (MethodInfo mi in mis)
            {
                Console.WriteLine();
                Console.WriteLine("現在のメソッド：" + mi.ToString());

                Console.WriteLine("親クラスの完全修飾名：" + mi.DeclaringType.FullName);
                Console.WriteLine("親クラスのメソッド：" + mi.ToString());

                // メソッドの一覧を表示します
                foreach (ParameterInfo pi in mi.GetParameters())
                {
                    Console.WriteLine("\t引数 {0}: {1}", pi.Name, pi.ParameterType.ToString());
                }
            }
        }

        public static void FindMethodOverrides(string asmPath, string typeName,
            string methodName, Type[] paramTypes)
        {
            Console.WriteLine(Environment.NewLine + "[型名] メソッド名：[" + typeName + "] " + methodName);

            Assembly asm = Assembly.LoadFrom(asmPath);

            // GetTypeメソッドは指定された型名が見つからない場合例外をスローすることに注意してください。
            // また、型名の大小文字は区別されません。
            Type asmType = asm.GetType(typeName, true, true);
            MethodInfo method = asmType.GetMethod(methodName, paramTypes);

            if (method != null)
            {
                MethodInfo baseDef = method.GetBaseDefinition();
                if (baseDef != method)
                {
                    Console.WriteLine("親クラスの完全修飾名：" +
                        baseDef.DeclaringType.FullName);
                    Console.WriteLine("親クラスのメソッド：" + baseDef.ToString());

                    bool foundMatch = false;

                    var match = from p in baseDef.GetParameters()
                                join op in paramTypes
                                    on p.ParameterType.UnderlyingSystemType equals op.UnderlyingSystemType
                                select p;

                    foundMatch = match.Any();

                    foreach (ParameterInfo pi in match)
                    {
                        // パラメータの一覧を表示して、メソッドを区別できるようにします
                        Console.WriteLine("\t引数 {0}: {1}",
                            pi.Name, pi.ParameterType.ToString());
                    }

                    // 検索対象のメソッドが見つかりました
                    if (foundMatch == true)
                    {
                        Console.WriteLine("条件に一致するメソッドが見つかりました。");
                    }
                }
            }
            Console.WriteLine();
        }
        #endregion

        #region "9.4 アセンブリ中からメンバを検索する"
        public static void FindMembersInAssembly()
        {
            string file = GetProcessPath();
            FindMemberInAssembly(file, "FindMembersInAssembly");
        }

        public static void FindMemberInAssembly(string asmPath, string memberName)
        {
            var members = from asm in Assembly.LoadFrom(asmPath).GetTypes()
                          from ms in asm.GetMember(memberName, MemberTypes.All,
                              BindingFlags.Public | BindingFlags.NonPublic |
                              BindingFlags.Static | BindingFlags.Instance)
                          select ms;

            foreach (MemberInfo member in members)
            {
                Console.WriteLine("{0} が見つかりました：クラス {2} の {1}",
                    member.MemberType,
                    member.ToString(),
                    member.DeclaringType.FullName);
            }
        }

        #endregion

        #region "9.5 アセンブリに含まれるネストされた型を検出して取得する"
        public static void ObtainNestedTypes()
        {
            string path = GetProcessPath();
            DisplayNestedTypes(path);
        }

        public static void DisplayNestedTypes(string asmPath)
        {
            var names = from t in Assembly.LoadFrom(asmPath).GetTypes()
                        from t2 in t.GetNestedTypes(BindingFlags.Instance |
                                    BindingFlags.Static |
                                    BindingFlags.Public |
                                    BindingFlags.NonPublic)
                        where !t2.IsEnum && !t2.IsInterface
                        select t2.FullName;

            foreach (string name in names)
            {
                Console.WriteLine(name);
            }
        }

        #endregion

        #region "9.6 型の継承階層を表示する"
        public static void DisplayInheritanceHierarchy()
        {
            DisplayInheritanceHierarchyType();
        }

        public static void DisplayInheritanceHierarchyType()
        {
            Process current = Process.GetCurrentProcess();
            // 現在のモジュールのパスを取得します
            string asmPath = current.MainModule.FileName;
            // 型を指定します
            DisplayInheritanceChain(asmPath, "CSharpRecipes.Reflection+DerivedOverrides");
            // アセンブリ中のすべての型を表示します
            DisplayInheritanceChain(asmPath);
        }

        public static void DisplayInheritanceChain(string asmPath)
        {
            Assembly asm = Assembly.LoadFrom(asmPath);
            var typeInfos = from Type type in asm.GetTypes()
                            select new
                            {
                                FullName = type.FullName,
                                BaseTypeDisplay = type.GetBaseTypeDisplay()
                            };

            foreach (var typeInfo in typeInfos)
            {
                // すべての基底クラスが再帰的に処理されています
                Console.WriteLine("派生クラス：" + typeInfo.FullName);
                Console.WriteLine("親クラスの一覧：" + typeInfo.BaseTypeDisplay);
                Console.WriteLine();
            }
        }

        public static void DisplayInheritanceChain(string asmPath, string baseType)
        {
            Assembly asm = Assembly.LoadFrom(asmPath);
            string typeDisplay = asm.GetType(baseType).GetBaseTypeDisplay();
            Console.WriteLine(typeDisplay);
        }
        #endregion

        #region "9.7 型のサブクラスを検索する"
        public static void FindSubclasses()
        {
            FindSubclassOfType();
        }

        public static void FindSubclassOfType()
        {
            Process current = Process.GetCurrentProcess();
            // 現在のモジュールのパスを取得します
            string asmPath = current.MainModule.FileName;
            Assembly asm = Assembly.LoadFrom(asmPath);
            Type type = Type.GetType("CSharpRecipes.Reflection+BaseOverrides");
            IEnumerable<Type> subClasses = asm.ListSubClassesForType(type);

            // 現在のクラスのサブクラスを表示します
            if (subClasses.Count() > 0)
            {
                Console.WriteLine("{0} にはサブクラスがあります：", type.FullName);
                foreach (Type t in subClasses)
                {
                    Console.WriteLine("\t{0}", t.FullName);
                }
            }
        }
        #endregion

        #region "9.8 アセンブリに含まれるすべてのシリアル化可能なクラスを検索する"
        public static void FindSerializableTypes()
        {
            FindSerializable();
        }

        public static void FindSerializable()
        {
            Process current = Process.GetCurrentProcess();
            // 現在のモジュールのパスを取得します
            string asmPath = current.MainModule.FileName;
            IEnumerable<string> typeNames = GetSerializableTypeNames(asmPath);
            // アセンブリに含まれるシリアル化可能なクラスを表示します
            if (typeNames.Count() > 0)
            {
                Console.WriteLine("{0} にはシリアル化可能なクラスがあります：", asmPath);
                foreach (string typeName in typeNames)
                {
                    Console.WriteLine("\t{0}", typeName);
                }
            }
        }

        public static IEnumerable<string> GetSerializableTypeNames(string asmPath)
        {
            Assembly asm = Assembly.LoadFrom(asmPath);
            return from type in asm.GetTypes()
                   where type.IsSerializable
                   select type.FullName;
        }
        #endregion

        #region "9.9 メンバの動的呼び出し"
        public static void DynamicInvocation()
        {
            TestDynamicInvocation();
        }

        public static void TestDynamicInvocation()
        {
            XDocument xdoc = XDocument.Load(@"..\..\SampleClassLibrary\SampleClassLibraryTests.xml");
            DynamicInvoke(xdoc, @"SampleClassLibrary.dll");
        }

        public static void DynamicInvoke(XDocument xdoc, string asmPath)
        {
            var test = from t in xdoc.Root.Elements("Test")
                       select new
                       {
                           typeName = (string)t.Attribute("className").Value,
                           methodName = (string)t.Attribute("methodName").Value,
                           argument = from p in t.Elements("Parameter")
                                      select new { arg = p.Value }
                       };

            // アセンブリを読み込みます
            Assembly asm = Assembly.LoadFrom(asmPath);

            foreach (var elem in test)
            {
                // 型に対して実際のTypeオブジェクトを作成します
                Type dynClassType = asm.GetType(elem.typeName, true, false);

                // インスタンスを作成して型が存在することを確認します
                object dynObj = Activator.CreateInstance(dynClassType);
                if (dynObj != null)
                {
                    // メソッドが存在して、MethodInfoオブジェクトが取得できることを確認します
                    MethodInfo invokedMethod = dynClassType.GetMethod(elem.methodName);
                    if (invokedMethod != null)
                    {
                        // メソッドを動的に呼び出すための引数リストを作成します
                        object[] arguments = new object[elem.argument.Count()];
                        int index = 0;

                        // 引数リストに各値を追加します
                        foreach (var arg in elem.argument)
                        {
                            // 引数の型を取得します
                            Type paramType =
                                invokedMethod.GetParameters()[index].ParameterType;

                            // 値をそれぞれの型に変換した後、リストに追加します
                            arguments[index] =
                                Convert.ChangeType(arg.arg, paramType);
                            index++;
                        }

                        // 引数を指定してメソッドを呼び出します
                        object retObj = invokedMethod.Invoke(dynObj, arguments);

                        Console.WriteLine("\t結果のオブジェクト：" + retObj);
                        Console.WriteLine("\t結果のオブジェクトの型名：" + retObj.GetType().FullName);
                    }
                }
            }
        }
        #endregion

        #region "9.10 クラスまたはメソッドがジェネリックかどうかを判別する"

        // ms-help://MS.VSCC.v80/MS.MSDNQTR.v80.en/MS.MSDN.v80/MS.NETDEVFX.v20.en/cpref/html/P_System_Type_ContainsGenericParameters.htm
        // ms-help://MS.VSCC.v80/MS.MSDNQTR.v80.en/MS.MSDN.v80/MS.VisualStudio.v80.en/dv_fxadvance/html/f93b03b0-1778-43fc-bc6d-35983d210e74.htm
        // ms-help://MS.VSCC.v80/MS.MSDNQTR.v80.en/MS.MSDN.v80/MS.VisualStudio.v80.en/dv_fxadvance/html/a0c3c0c5-8178-4673-8107-9fdde780a669.htm


        public static void TestIsGeneric()
        {
            string path = GetProcessPath();

            // アセンブリを読み込みます
            Assembly asm = Assembly.LoadFrom(path);
            Console.WriteLine("アセンブリ {0}:", path);

            //foreach (Type tp in asm.GetTypes())
            //{
            //    if (tp.FullName.StartsWith("CSharpRecipes.DataStructsAndAlgorithms+"))
            //        Console.WriteLine("Name == " + tp.FullName);
            //}
            //Console.WriteLine();
            //Console.WriteLine();

            // Typeオブジェクトを取得します
            Type t = asm.GetType("CSharpRecipes.DataStructsAndAlgorithms+PriorityQueue`1");

            bool genericType = IsGenericType(t);

            bool genericMethod = false;
            foreach (MethodInfo mi in t.GetMethods())
                genericMethod = IsGenericMethod(mi);
            GetGenericMemberInfo(t);
        }

        public static bool IsGenericType(Type type)
        {
            Console.WriteLine("IsGenericType: {0}", type.IsGenericType);

            return type.IsGenericType;
        }

        public static bool IsGenericMethod(MethodInfo mi)
        {
            Console.WriteLine("Name: {0}", mi.Name);
            Console.WriteLine("IsGenericMethod: {0}", mi.IsGenericMethod);

            return mi.IsGenericMethod;
        }

        public static void GetGenericMemberInfo(Type t)
        {

            Console.WriteLine("HasGenericArguments: {0}", t.IsGenericType);
            Console.WriteLine("IsGenericTypeDefinition: {0}", t.IsGenericTypeDefinition);

            Console.WriteLine("FullName: {0}", t.FullName);
            Console.WriteLine("ContainsGenericParameters: {0}", t.ContainsGenericParameters);
            Console.WriteLine("IsGenericParameter: {0}", t.IsGenericParameter);
            if (t.IsGenericParameter)
            {
                Console.WriteLine("DeclaringMethod.Name: {0}", t.DeclaringMethod.Name);
                Console.WriteLine("GenericParameterAttributes: {0}", t.GenericParameterAttributes);
                Console.WriteLine("GenericParameterPosition: {0}", t.GenericParameterPosition);
                foreach (Type innerType in t.GetGenericParameterConstraints())
                    Console.WriteLine("\tGetGenericParameterConstraints().FullName: {0}", innerType.FullName);
            }
            foreach (Type innerType in t.GetGenericArguments())
                Console.WriteLine("\tGetGenericArguments().FullName: {0}", innerType.FullName);
            Console.WriteLine("GetGenericTypeDefinition().FullName: {0}", t.GetGenericTypeDefinition().FullName);
            Console.WriteLine("Name: {0}", t.Name);
        }
        #endregion

        #region "9.11 ローカル変数に関する情報を参照する"
        public static void TestGetLocalVars()
        {
            string file = GetProcessPath();

            // CSharpRecipes.Reflection.GetLocalVarsメソッドの中にある
            // すべてのローカル変数についての情報を取得します
            IList<LocalVariableInfo> vars =
                GetLocalVars(file, "CSharpRecipes.Reflection", "GetLocalVars");
        }

        public static IList<LocalVariableInfo> GetLocalVars(string asmPath, string typeName, string methodName)
        {
            Assembly asm = Assembly.LoadFrom(asmPath);
            Type asmType = asm.GetType(typeName);
            MethodInfo mi = asmType.GetMethod(methodName);
            MethodBody mb = mi.GetMethodBody();

            IList<LocalVariableInfo> vars = mb.LocalVariables;

            // それぞれのローカル変数についての情報を表示します
            foreach (LocalVariableInfo lvi in vars)
            {
                Console.WriteLine("IsPinned: " + lvi.IsPinned);
                Console.WriteLine("LocalIndex: " + lvi.LocalIndex);
                Console.WriteLine("LocalType.Module: " + lvi.LocalType.Module);
                Console.WriteLine("LocalType.FullName: " + lvi.LocalType.FullName);
                Console.WriteLine("ToString(): " + lvi.ToString());
            }

            return vars;
        }
        #endregion

        #region "9.12 ジェネリック型の作成"

        public static void TestCreateMultiMap()
        {
            Assembly asm = Assembly.LoadFrom("CSharpRecipes.exe");
            CreateMultiMap(asm);
        }

        public static void CreateMultiMap(Assembly asm)
        {
            // 作成する型を表すTypeを取得します
            Type typeToConstruct = asm.GetType(
                "CSharpRecipes.DataStructsAndAlgorithms+MultiMap`2");
            // 型パラメータに指定するための型を取得します
            Type[] typeArguments = { typeof(int), typeof(string) };
            // 型パラメータをジェネリック型に関連づけます
            Type newType = typeToConstruct.MakeGenericType(typeArguments);
            // 型のインスタンスを作成します
            DataStructsAndAlgorithms.MultiMap<int, string> mm =
                (DataStructsAndAlgorithms.MultiMap<int, string>)Activator.CreateInstance(newType);

            // 作成した型をテストします
            Console.WriteLine("Count == " + mm.Count);
            mm.Add(1, "test1");
            Console.WriteLine("Count == " + mm.Count);
        }
        #endregion

        #region "9.B 付録 リフレクション経由でMSILにアクセスする
        public static void TestGetIL()
        {
            string file = GetProcessPath();

            byte[] IL = GetIL(file, "CSharpRecipes.Reflection", "DynamicInvocation");

            foreach (byte b in IL)
            {
                Console.WriteLine(b.ToString("X2"));
            }

            ResolveMember(IL);
        }

        public static byte[] GetIL(string asmPath, string typeName, string methodName)
        {
            Assembly asm = Assembly.LoadFrom(asmPath);
            Type asmType = asm.GetType(typeName);
            MethodInfo mi = asmType.GetMethod(methodName);

            foreach (MethodInfo testMI in asmType.GetMethods())
            {
                Console.WriteLine("\t-->  " + testMI.Name + "\t" + testMI.MetadataToken.ToString("X"));
            }

            MethodBody mb = mi.GetMethodBody();

            byte[] IL = mb.GetILAsByteArray();

            return IL;
        }

        public static void ResolveMember(byte[] IL)
        {
            int currByte = 0;

            //Process current = Process.GetCurrentProcess();

            while (currByte < IL.Length)
            {
                switch (IL[currByte])
                {
                    // call命令
                    case 0x28:
                        {
                            // 続く4命令を取得します
                            int token = GetTokenFromBytes(IL[++currByte], IL[++currByte], IL[++currByte], IL[++currByte]);

                            foreach (Module m in Assembly.GetExecutingAssembly().GetLoadedModules())
                            {
                                try
                                {
                                    MemberInfo mi = m.ResolveMember(token);
                                    Console.WriteLine("解決された名称：" + mi.Name);
                                }
                                catch (ArgumentException)
                                {
                                    // 現在のモジュールに見つかりません
                                }
                            }
                            break;
                        }
                }
                currByte++;
            }
        }

        public static int GetTokenFromBytes(byte b0, byte b1, byte b2, byte b3)
        {
            Console.WriteLine("\tb0 == " + b0.ToString("X"));
            Console.WriteLine("\tb1 == " + b1.ToString("X"));
            Console.WriteLine("\tb2 == " + b2.ToString("X"));
            Console.WriteLine("\tb3 == " + b3.ToString("X"));

            string token = b0.ToString("X");
            token = "0" + b1.ToString("X") + token;
            token = "0" + b2.ToString("X") + token;
            token = "0" + b3.ToString("X") + token;

            Console.WriteLine("MetaData Token(X): " + token.ToString());

            int intToken = int.Parse(token, System.Globalization.NumberStyles.HexNumber);

            Console.WriteLine("intToken == " + intToken.ToString("X"));

            return intToken;
        }


        /*
            // 以下のアドレスは実行環境によって異なる場合があります

                -->  DynamicInvocation  60005BB
                -->  TestDynamicInvocation      60005BC

                00
                28  BC  05  00  06
                00
                2A 
		  
		 
                ILDASM
                00  NOP
                28  (06)0005BC  call void CSharpRecipes.Reflection::TestDynamicInvocation()
                00  NOP
                2A  RET
        */

        /* ILの出力例：
            00   |                   nop
            02   |                   ldarg.0
            7B   | (0A)0005DF        ldfld
            03   |                   ldarg.1
            6F   | (0A)0001E1        callvirt
            00   |                   nop
            02   |                   ldarg.0
            7B   | (0A)0005DF        ldfld
            02   |                   ldarg.0
            7B   | (0A)0005E0        ldfld
            6F   | (0A)0005E8        callvirt
            00   |                   nop
            2A   |                   ret
		  
          上のメソッドに対応するIL出力
            00
            02
            7B   | E0	05	00	0A
            03
            6F   |	E1	01	00	0A
            0
            2
            7B   |	E0	05	00	0A
            2
            7B   | 	E1	05	00	0A
            6F   | E9	05	00	0A
            0
            2A
        */
        #endregion
    }
}
