﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Runtime.Remoting.Messaging;
using System.Threading;

namespace CSharpRecipes
{
    public class ThreadingSync
    {
        #region "11.1 Threadごとにstaticフィールドを作成する"
        public static void PerThreadStatic()
        {
            TestStaticField();
        }

        public class Foo
        {
            [ThreadStaticAttribute()]
            public static string bar = "初期化文字列";
        }

        private static void TestStaticField()
        {
            ThreadStaticField.DisplayStaticFieldValue();

            Thread newStaticFieldThread =
                new Thread(ThreadStaticField.DisplayStaticFieldValue);

            newStaticFieldThread.Start();

            ThreadStaticField.DisplayStaticFieldValue();
        }

        private class ThreadStaticField
        {
            [ThreadStaticAttribute()]
            public static string bar = "初期化文字列";

            public static void DisplayStaticFieldValue()
            {
                string msg =
                    string.Format("{0} には次の値のstaticフィールドがあります：{1}",
                        Thread.CurrentThread.GetHashCode(),
                        ThreadStaticField.bar);
                Console.WriteLine(msg);
            }
        }


        #endregion

        #region "11.2 クラスメンバへのスレッドセーフなアクセスを実現する"
        public static void ThreadSafeAccess()
        {
            DeadLock deadLock = new DeadLock();
            lock (deadLock)
            {
                Thread thread = new Thread(deadLock.Method1);
                thread.Start();

                // 何らかの処理を実行します。
            }

            int num = 0;
            if (Monitor.TryEnter(MonitorMethodAccess.SyncRoot, 250))
            {
                MonitorMethodAccess.ModifyNumericField(10);
                num = MonitorMethodAccess.ReadNumericField();
                Monitor.Exit(MonitorMethodAccess.SyncRoot);
            }
            Console.WriteLine(num);

        }

        public static class NoSafeMemberAccess
        {
            private static int numericField = 1;

            public static void IncrementNumericField()
            {
                ++numericField;
            }

            public static void ModifyNumericField(int newValue)
            {
                numericField = newValue;
            }

            public static int ReadNumericField()
            {
                return numericField;
            }
        }

        public static class SaferMemberAccess
        {
            private static int numericField = 1;
            private static object syncObj = new object();

            public static void IncrementNumericField()
            {
                lock (syncObj)
                {
                    ++numericField;
                }
            }

            public static void ModifyNumericField(int newValue)
            {
                lock (syncObj)
                {
                    numericField = newValue;
                }
            }

            public static int ReadNumericField()
            {
                lock (syncObj)
                {
                    return numericField;
                }
            }
        }

        public class DeadLock
        {
            private object syncObj = new object();

            public void Method1()
            {
                lock (syncObj)
                {
                    // 何らかの処理を実行します。
                }
            }
        }

        public static class MonitorMethodAccess
        {
            private static int numericField = 1;
            private static object syncObj = new object();

            public static object SyncRoot
            {
                get { return syncObj; }
            }

            public static void IncrementNumericField()
            {
                if (Monitor.TryEnter(syncObj, 250))
                {
                    try
                    {
                        ++numericField;
                    }
                    finally
                    {
                        Monitor.Exit(syncObj);
                    }
                }
            }

            public static void ModifyNumericField(int newValue)
            {
                if (Monitor.TryEnter(syncObj, 250))
                {
                    try
                    {
                        numericField = newValue;
                    }
                    finally
                    {
                        Monitor.Exit(syncObj);
                    }
                }
            }

            public static int ReadNumericField()
            {
                if (Monitor.TryEnter(syncObj, 250))
                {
                    try
                    {
                        return numericField;
                    }
                    finally
                    {
                        Monitor.Exit(syncObj);
                    }
                }

                return -1;
            }
            [MethodImpl(MethodImplOptions.Synchronized)]
            public static void MySynchronizedMethod()
            {
            }

        }


        #endregion

        #region "11.3 暗黙的なスレッド終了を回避する"
        public static void PreventSilentTermination()
        {
            MainThread mt = new MainThread();
            mt.CreateNewThread();
        }

        public class MainThread
        {
            public void CreateNewThread()
            {
                // 現在の作業を行うために新しいスレッドを作成します
                Thread newWorkerThread = new Thread(Worker.DoWork);
                newWorkerThread.Start();
            }
        }

        public class Worker
        {
            // 現在の作業をするためにThreadStartデリゲート経由で呼ばれるメソッドです
            public static void DoWork()
            {
                try
                {
                    // スレッドの作業はここで行います
                    throw new Exception("ボン！");
                }
                catch (Exception e)
                {
                    // スレッドの例外はここで処理します
                    Console.WriteLine(e.ToString());
                    // 例外を再スローしてはいけません
                }
                finally
                {
                    // スレッドの終了処理はここで行います
                }
            }
        }
        #endregion

        #region "11.4 非同期デリゲートの完了通知を受け取る"
        public static void CompletionAsyncDelegate()
        {
            AsyncAction2 aa2 = new AsyncAction2();
            aa2.CallbackAsyncDelegate();
        }

        public delegate int AsyncInvoke();

        public class TestAsyncInvoke
        {
            public static int Method1()
            {
                Console.WriteLine("スレッド{0}でMethod1が呼び出されました。",
                    Thread.CurrentThread.ManagedThreadId);
                return 1;
            }
        }

        public class AsyncAction2
        {
            public void CallbackAsyncDelegate()
            {
                AsyncCallback callBack = DelegateCallback;

                AsyncInvoke method1 = TestAsyncInvoke.Method1;
                Console.WriteLine("スレッド{0}でBeginInvokeを呼び出します。",
                    Thread.CurrentThread.ManagedThreadId);
                IAsyncResult asyncResult = method1.BeginInvoke(callBack, method1);

                // ここではポーリングしたりWaitOneメソッドを使用する必要がないため、
                // 呼び出し元のメソッドに戻ります。
                return;
            }

            private static void DelegateCallback(IAsyncResult iresult)
            {
                Console.WriteLine("スレッド{0}でコールバックが呼ばれました。",
                    Thread.CurrentThread.ManagedThreadId);
                AsyncResult asyncResult = (AsyncResult)iresult;
                AsyncInvoke method1 = (AsyncInvoke)asyncResult.AsyncDelegate;

                int retVal = method1.EndInvoke(asyncResult);
                Console.WriteLine("デリゲートの返り値(コールバック): " + retVal);
            }
        }

        public delegate int AsyncInvoke2();

        public class TestAsyncInvoke2
        {
            public static int Method1()
            {
                Console.WriteLine("スレッド{0}でMethod1が呼び出されました。",
                    Thread.CurrentThread.ManagedThreadId);
                return 1;
            }
        }

        #endregion

        #region "11.5 スレッド固有のデータを非公開で保持する"
        public static void StoreThreadDataPrivately()
        {
            HandleClass.Run();
        }

        public class ApplicationData
        {
            // アプリケーションのデータがここに格納されます
            public int Data { get; set; }
        }

        public class HandleClass
        {
            public static void Run()
            {
                // データのインスタンスを作成し、名前を付けたデータスロットに保持します
                ApplicationData appData = new ApplicationData();
                Thread.SetData(Thread.GetNamedDataSlot("appDataSlot"), appData);

                // このデータを使用する他のメソッドを呼び出します。
                HandleClass.MethodB();

                // 終了後にデータスロットを解放します
                Thread.FreeNamedDataSlot("appDataSlot");
            }

            public static void MethodB()
            {
                // データのインスタンスを名前付きデータスロットから取得します
                ApplicationData storedAppData = (ApplicationData)Thread.GetData(
                    Thread.GetNamedDataSlot("appDataSlot"));

                // ApplicationDataを更新します

                // データの変更後、変更をデータスロットに保存しなおします
                Thread.SetData(Thread.GetNamedDataSlot("appDataSlot"),
                    storedAppData);

                // このデータを使用する他のメソッドを呼び出します
                HandleClass.MethodC();
            }

            public static void MethodC()
            {
                // データのインスタンスを名前付きデータスロットから取得します
                ApplicationData storedAppData =
                    (ApplicationData)Thread.GetData(Thread.GetNamedDataSlot("appDataSlot"));

                // ApplicationDataを変更します

                // データの変更後、変更をデータスロットに保存しなおします
                Thread.SetData(Thread.GetNamedDataSlot("appDataSlot"), storedAppData);
            }
        }

        #endregion

        #region "11.6 セマフォを利用した複数スレッドからのリソースアクセス許可"
        // Xbox360には、4人が同時にプレイできるように4つコントローラのポートがあります。
        // 最初はXbox360が起動してゲームをロードするまでプレイヤー達が並んで待っているため、
        // アクセス数の最大値を4、初期値を0とします。

        public class Xbox360Player
        {
            public class PlayerInfo
            {
                public ManualResetEvent Dead { get; set; }
                public string Name { get; set; }
            }

            // プレイヤーの死に様
            private static string[] _deaths = new string[7]{"戦死した",
																"窒息死した",
																"自爆した",
																"生け捕りにされた",
																"墜落死した",
																"毒殺された",
																"グレネードに被弾した",
																};

            /// <summary>
            /// 別スレッドで実行されるメソッドです
            /// </summary>
            /// <param name="info">Xboxへの参照を保持しているXbox360Player.Data要素</param>
            public static void JoinIn(object info)
            {
                // セマフォを扱えるよう、名前で指定してセマフォを開きます。
                using (Semaphore Xbox360 = Semaphore.OpenExisting("Xbox360"))
                {

                    // データオブジェクトの取得
                    PlayerInfo player = (PlayerInfo)info;

                    // 各プレイヤーがXbox360に遊びたい旨を通知します。
                    Console.WriteLine("{0}はゲームを待っています!", player.Name);

                    // コントローラを握る権利をくれるまでXbox360(セマフォ)を待ちます
                    Xbox360.WaitOne();

                    // Xbox360がプレイヤーを選びました! 
                    // (つまりセマフォがリソースへのアクセスを許可しました...)
                    Console.WriteLine("{0}がプレイ可能になりました。 " +
                        "Welcome to your doom {0}. >:)", player.Name);

                    // プレイヤーが生き延びる時間を決定するランダム値を算出します。
                    System.Random rand = new Random(500);
                    int timeTillDeath = rand.Next(100, 1000);

                    // ゲームオーバーになるまでプレイヤーが忙しい状態をシミュレートします。
                    Thread.Sleep(timeTillDeath);

                    // プレイヤーの死に様を決定します。
                    rand = new Random();
                    int deathIndex = rand.Next(6);

                    // プレイヤーの死を通知します。
                    Console.WriteLine("{0}は{1}ため、次のプレイヤーの番になります。",
                        player.Name, _deaths[deathIndex]);

                    // 自分以外のすべてのポートが空いていた場合は、
                    // プレイヤー全員が遊んでゲームオーバーになったことを示します。
                    int semaphoreCount = Xbox360.Release();
                    if (semaphoreCount == 3)
                    {
                        Console.WriteLine("お疲れ様でした。ゲームを終了します。");
                        // プレイヤーのDeadイベントを登録します。
                        player.Dead.Set();
                        // セマフォを閉じます。
                        Xbox360.Close();
                    }
                }
            }
        }

        public class Halo3Session
        {
            // 制限されたリソースプールを表すSemaphoreオブジェクト
            private static Semaphore _Xbox360;

            public static void Play()
            {
                // Xbox360には、4人が同時にプレイできるように4つコントローラのポートがあります。
                // 最初はXbox360が起動してゲームをロードするまでプレイヤー達が並んで待っているため、
                // アクセス数の最大値を4、0を初期値とします。
                //
                using (_Xbox360 = new Semaphore(0, 4, "Xbox360"))
                {
                    using (ManualResetEvent GameOver = new ManualResetEvent(false))
                    {
                        //
                        // 9人のプレイヤーがログインします。
                        //
                        List<Xbox360Player.PlayerInfo> players = new List<Xbox360Player.PlayerInfo>() {
                            new Xbox360Player.PlayerInfo { Name="Igor", Dead=GameOver},
                            new Xbox360Player.PlayerInfo { Name="AxeMan", Dead=GameOver},
                            new Xbox360Player.PlayerInfo { Name="Dr. Death",Dead=GameOver},
                            new Xbox360Player.PlayerInfo { Name="HaPpyCaMpEr",Dead=GameOver},
                            new Xbox360Player.PlayerInfo { Name="Executioner",Dead=GameOver},
                            new Xbox360Player.PlayerInfo { Name="FragMan",Dead=GameOver},
                            new Xbox360Player.PlayerInfo { Name="Beatdown",Dead=GameOver},
                            new Xbox360Player.PlayerInfo { Name="Stoney",Dead=GameOver},
                            new Xbox360Player.PlayerInfo { Name="Pwned",Dead=GameOver}
                        };

                        foreach (Xbox360Player.PlayerInfo player in players)
                        {
                            Thread t = new Thread(Xbox360Player.JoinIn);

                            // スレッドにプレイヤー名を付けます。
                            t.Name = player.Name;
                            // プレイヤーがHalo3をプレイするため列に並びます。
                            t.Start(player);
                        }

                        // Xbox360が起動してHalo3をロードします(3秒)。
                        Console.WriteLine("Xbox360を初期化しています...");
                        Thread.Sleep(3000);
                        Console.WriteLine("Halo3がロードされました。現在、あと4人がプレイできます...");

                        // Xbox360はセマフォのカウント全部を保持しています。
                        // Release(4)によって4つのスロットが空けられ、待っているプレイヤーが
                        // 同時に4人までXbox360(セマフォ)に接続できるようになります。
                        //
                        _Xbox360.Release(4);

                        // プレイヤーのゲームオーバーを待ちます...
                        GameOver.WaitOne();
                    }
                }
            }
        }
        #endregion

        #region "11.7 Mutexを使用した複数プロセスの同期"
        // MutexFunプロジェクトを参照してください
        #endregion

        #region "11.8 イベントを使用して複数スレッドを協調する"
        public static void TestResetEvent()
        {
            // 夕食の時間ですが、料理人は1度に1品しか料理を出すことができません。
            Cook Mel = new Cook();

            // ウェイトレスを5人用意してそれぞれがオーダーを受けます
            for (int i = 0; i < 5; i++)
            {
                Thread t = new Thread(Waitress.PlaceOrder);
                // ウェイトレスはオーダーを報告した後、再びオーダーを待ちます
                t.Start(Cook.OrderReady);
            }

            // 全員の準備が整いました
            for (int i = 0; i < 5; i++)
            {
                // ウェイトレスを待機させます
                Thread.Sleep(2000);
                // 準備OKです。次のウェイトレスを呼び出します。
                Mel.CallWaitress();
            }
        }

        public class Cook
        {
            public static AutoResetEvent OrderReady = new AutoResetEvent(false);

            public void CallWaitress()
            {
                // AutoResetEventのSetメソッドを実行しますが、ManualResetEventとは異なり、
                // 再度イベントを発生させるためにResetメソッドを実行する必要はありません。
                // このイベントをセットするとPlaceOrderメソッド中で
                // 待機状態になっているウェイトレスへ通知されます。
                OrderReady.Set();
            }
        }

        public class Waitress
        {
            public static void PlaceOrder(object signal)
            {
                // AutoResetEventにキャストしてオーダーを受ける準備が
                // 整うまで待機できるようにします
                AutoResetEvent OrderReady = (AutoResetEvent)signal;
                // オーダーを待ちます...
                OrderReady.WaitOne();
                // オーダーを受けました...
                Console.WriteLine("ウェイトレスはオーダーを受けました！");
            }
        }

        #endregion

        #region "11.9 イベントの命名権を取得する"
        public static void TestManualNamedEvent()
        {
            // 名前付きのマニュアルリセットイベントを作成します
            EventWaitHandle ewhSuperBowl =
                new EventWaitHandle(false, // 初期状態は非シグナル状態とします
                                    EventResetMode.ManualReset,
                                    @"Champs");

            // イベントをリッスンするスレッドを3つ動かします
            for (int i = 0; i < 3; i++)
            {
                Thread t = new Thread(ManualFan);
                // 熱心なファンが待ち構えています...
                t.Name = "Fan " + i;
                t.Start();
            }
            // ゲームを開始します
            Thread.Sleep(10000);
            // 告知を行います
            Console.WriteLine("Patriotsがスーパーボウルを制覇しました！");
            // すべてのファンにシグナルを送ります
            ewhSuperBowl.Set();
            // イベントをクローズします
            ewhSuperBowl.Close();

        }

        public static void TestAutoNamedEvent()
        {
            // 名前付きのオートリセットイベントを作成します
            EventWaitHandle ewhSuperBowl =
                new EventWaitHandle(false, // 初期状態は非シグナル状態とします
                                    EventResetMode.AutoReset,
                                    @"Champs");

            // イベントをリッスンするスレッドを3つ動かします
            for (int i = 0; i < 3; i++)
            {
                Thread t = new Thread(AutoFan, i);
                // 熱心なファンが待ち構えています...
                t.Name = "Fan " + i;
                t.Start();
            }
            // ゲームを開始します
            Thread.Sleep(10000);
            // 告知を行います
            Console.WriteLine("Patriotsがスーパーボウルを制覇しました！");
            // 毎回1人のファンにシグナルを送ります
            for (int i = 0; i < 3; i++)
            {
                Console.WriteLine("ファンに通知中");
                ewhSuperBowl.Set();
            }
            // イベントをクローズします
            ewhSuperBowl.Close();
        }

        public static void ManualFan()
        {
            // 名前付きイベントをオープンします
            EventWaitHandle ewhSuperBowl =
                new EventWaitHandle(false,
                                    EventResetMode.ManualReset,
                                    @"Champs");

            // シグナルを待機します
            ewhSuperBowl.WaitOne();
            // 告知を受けて叫びます
            Console.WriteLine("\"よくやった！\" スレッド名：{0}", Thread.CurrentThread.Name);
            // イベントをクローズします
            ewhSuperBowl.Close();
        }

        public static void AutoFan()
        {
            // 名前付きイベントをオープンします
            EventWaitHandle ewhSuperBowl =
                new EventWaitHandle(false,
                                    EventResetMode.ManualReset,
                                    @"Champs");

            // シグナルを待機します
            ewhSuperBowl.WaitOne();
            // 告知を受けて叫びます
            Console.WriteLine("\"やった！\" スレッド名：{0}", Thread.CurrentThread.Name);
            // イベントをクローズします
            ewhSuperBowl.Close();
        }
        #endregion

        #region "11.10 スレッド間で処理を不可分に実行する"
        public static void TestInterlocked()
        {
            int i = 0;
            long l = 0;
            Interlocked.Increment(ref i); // i = 1
            Interlocked.Decrement(ref i); // i = 0
            Interlocked.Increment(ref l); // l = 1
            Interlocked.Decrement(ref i); // l = 0

            Interlocked.Add(ref i, 10); // i = 10;
            Interlocked.Add(ref l, 100); // l = 100;

            string name = "Mr. Ed";
            Interlocked.Exchange(ref name, "Barney");

            double runningTotal = 0.0;
            double startingTotal = 0.0;
            double calc = 0.0;
            for (i = 0; i < 10; i++)
            {
                do
                {
                    // 元の合計値を保存します
                    startingTotal = runningTotal;
                    // 複雑な計算を行います
                    calc = runningTotal + i * Math.PI * 2 / Math.PI;
                }
                // runningTotalの値が変更されていないことを確認して
                // 変更されていなければ新しい値に置き換えます。
                // 変更されていた場合は現在の値が取得できるまでループを繰り返します。
                while (startingTotal !=
                    Interlocked.CompareExchange(
                        ref runningTotal, calc, startingTotal));
            }

            // それほど信頼のないオークション
            Auction Shadys = new Auction("Shady's");
            // 信頼のあるオークション
            Auction UpperCrust = new Auction("UpperCrust");

            // 「無鉄砲な」小口入札者と「手堅い」小口入札者
            // および「無鉄砲な」大口入札者と「手堅い」大口入札者をそれぞれ3人用意します
            Thread[] handles = new Thread[12];
            for (i = 0; i < 12; i++)
            {
                Thread t = null;
                switch (i)
                {
                    // 小口の無鉄砲な入札者
                    case 0:
                    case 1:
                    case 2:
                        {
                            t = new Thread(new ThreadStart(Shadys.Bid));
                        }
                        break;
                    // 小口の手堅い入札者
                    case 3:
                    case 4:
                    case 5:
                        {
                            t = new Thread(new ThreadStart(UpperCrust.BidSafe));
                        }
                        break;
                    // 大口の無鉄砲な入札者
                    case 6:
                    case 7:
                    case 8:
                        {
                            t = new Thread(new ThreadStart(Shadys.BigBid));
                        }
                        break;
                    // 大口の手堅い入札者
                    case 9:
                    case 10:
                    case 11:
                        {
                            t = new Thread(new ThreadStart(UpperCrust.BigBidSafe));
                        }
                        break;
                }
                // 後で使用するためにスレッド間で参照可能にします
                handles[i] = t;

                // スレッドを開始
                t.Start();
            }

            // それぞれのスレッドを連結して、すべての処理が完了するまで待機します
            for (i = 0; i < 12; i++)
                handles[i].Join();

            Console.WriteLine(
                "Shadyオークションでは総額 {1} ドル、計 {0} 口の入札がありました",
                Shadys.BidCount, Shadys.BidTotal);
            Console.WriteLine(
                "UpperCrustオークションでは総額 {1} ドル、計 {0} 口の入札がありました",
                UpperCrust.BidCount, UpperCrust.BidTotal);
        }

        /// <summary>
        /// オークションをシミュレートするクラス
        /// </summary>
        public class Auction
        {
            #region Public Data
            private int _bidCount;
            private long _bidTotal;

            // 入札件数
            public int BidCount
            {
                get { return _bidCount; }
                set { _bidCount = value; }
            }
            // 総入札件数
            public long BidTotal
            {
                get { return _bidTotal; }
                set { _bidTotal = value; }
            }
            #endregion

            #region Private Data
            // オークションの名前
            private string _name;
            #endregion

            #region CTOR
            /// <summary>
            /// オークション名をConstructor that takes the name
            /// </summary>
            /// <param name="name">オークション名</param>
            public Auction(string name)
            {
                _name = name;
            }
            #endregion

            #region Private Methods
            /// <summary>
            /// 0から2秒の間でランダムな時間待機します
            /// </summary>
            private void Wait()
            {
                // 何らかの作業を行うためにランダムな時間待機します
                Random r = new Random();
                int milli = r.Next(2000);
                Thread.Sleep(milli);
            }
            #endregion

            #region Public Thread Methods
            public void Bid()
            {
                // 何らかの作業を行うためにランダムな時間待機します
                Wait();

                // スレッドセーフかどうかを確認せずに入札件数を1追加します
                BidCount++;
                BidTotal++;

                // 入札があったことを通知します
                Console.WriteLine("{0} にてスレッド {1} から入札がありました。",
                    _name, Thread.CurrentThread.ManagedThreadId);
            }

            public void BidSafe()
            {
                // 何らかの作業を行うためにランダムな時間待機します
                Wait();

                // スレッドセーフを考慮して入札件数を1追加します
                Interlocked.Increment(ref _bidCount);
                Interlocked.Increment(ref _bidTotal);

                // 入札があったことを通知します
                Console.WriteLine("{0} にてスレッド {1} からスレッドセーフな入札がありました。",
                    _name, Thread.CurrentThread.ManagedThreadId);
            }

            public void BigBid()
            {
                // 何らかの作業を行うためにランダムな時間待機します
                Wait();

                // スレッドセーフかどうかを確認せずに入札件数を10追加します
                BidCount += 1;
                BidTotal += 10;

                // 入札があったことを通知します
                Console.WriteLine("{0} にてスレッド {1} から大口の入札がありました。",
                    _name, Thread.CurrentThread.ManagedThreadId);
            }

            public void BigBidSafe()
            {
                // 何らかの作業を行うためにランダムな時間待機します
                Wait();

                // スレッドセーフを考慮して入札件数を10追加します
                Interlocked.Add(ref _bidCount, 1);
                Interlocked.Add(ref _bidTotal, 10);

                // 入札があったことを通知します
                Console.WriteLine("{0} にてスレッド {1} からスレッドセーフな大口の入札がありました。",
                    _name, Thread.CurrentThread.ManagedThreadId);
            }
            #endregion
        }
        #endregion

        #region "11.11 読み取り主体のアクセスを効率的に実行する"

        //static Developer _dev = new Developer(15);
        static Developer _dev = null;
        static bool _end = false;

        /// <summary>
        /// </summary>
        public static void TestReaderWriterLockSlim()
        {
            LaunchTeam(_dev);
            Thread.Sleep(10000);
        }

        private static void LaunchTeam(Developer dev)
        {
            LaunchManager("CTO", dev);
            LaunchManager("ディレクター", dev);
            LaunchManager("プロジェクトマネージャ", dev);
            LaunchDependent("製品マネージャ", dev);
            LaunchDependent("テストエンジニア", dev);
            LaunchDependent("技術スタッフ", dev);
            LaunchDependent("作業スタッフ", dev);
            LaunchDependent("サポートスタッフ", dev);
        }

        public class TaskInfo
        {
            private Developer _dev;
            public string Name { get; set; }
            public Developer Developer
            {
                get { return _dev; }
                set { _dev = value; }
            }
        }

        private static void LaunchManager(string name, Developer dev)
        {
            ThreadPool.QueueUserWorkItem(
                new WaitCallback(CreateManagerOnThread),
                new TaskInfo() { Name = name, Developer = dev });
        }

        private static void LaunchDependent(string name, Developer dev)
        {
            ThreadPool.QueueUserWorkItem(
                new WaitCallback(CreateDependentOnThread),
                new TaskInfo() { Name = name, Developer = dev });
        }

        private static void CreateManagerOnThread(object objInfo)
        {
            TaskInfo taskInfo = (TaskInfo)objInfo;
            Console.WriteLine("タスク " + taskInfo.Name + " をプロジェクトに追加しました...");
            TaskManager mgr = new TaskManager(taskInfo.Name, taskInfo.Developer);
        }

        private static void CreateDependentOnThread(object objInfo)
        {
            TaskInfo taskInfo = (TaskInfo)objInfo;
            Console.WriteLine("タスク " + taskInfo.Name + " をプロジェクトに追加しました...");
            TaskDependent dep = new TaskDependent(taskInfo.Name, taskInfo.Developer);
        }

        public class Task
        {
            public Task(string name)
            {
                Name = name;
            }

            public string Name { get; set; }
            public int Priority { get; set; }
            public bool Status { get; set; }

            public override string ToString()
            {
                return this.Name;
            }

            public override bool Equals(object obj)
            {
                Task task = obj as Task;
                if (task != null)
                    return this.Name == task.Name;
                return false;
            }

            public override int GetHashCode()
            {
                return this.Name.GetHashCode();
            }
        }

        public class Developer
        {
            /// <summary>
            /// タスクを表すオブジェクトです
            /// </summary>
            private List<Task> _tasks = new List<Task>();
            private ReaderWriterLockSlim _rwlSlim = new ReaderWriterLockSlim();
            private System.Threading.Timer _timer;
            private int _maxTasks;

            public Developer(int maxTasks)
            {
                // Developerが受け入れ可能な最大タスク数を設定します
                _maxTasks = maxTasks;
                // それぞれのタスクは1/4秒毎に終了します
                _timer = new Timer(new TimerCallback(DoWork), null, 1000, 250);
            }

            ~Developer()
            {
                _timer.Dispose();
            }


            // タスクを実行します
            protected void DoWork(Object stateInfo)
            {
                ExecuteTask();
                try
                {
                    _rwlSlim.EnterWriteLock();
                    // タスクをすべて完了したならば休暇を取るチャンスです！
                    if (_tasks.Count == 0)
                    {
                        _end = true;
                        Console.WriteLine("Developerはすべてのタスクを完了しました。休暇を取るチャンスです！");
                        return;
                    }

                    if (!_end)
                    {
                        // タスクが多すぎる場合は辞職してしまいます
                        if (_tasks.Count > _maxTasks)
                        {
                            // 未完了のタスクの数を取得します
                            var query = from t in _tasks
                                        where t.Status == false
                                        select t;
                            int unfinishedTaskCount = query.Count<Task>();

                            _end = true;
                            Console.WriteLine("Developerはあまりに多くのタスクを抱えすぎたため、辞職しました！" +
                                unfinishedTaskCount + " 個のタスクが未完了です。");
                        }
                    }
                    else
                        _timer.Dispose();
                }
                finally
                {
                    _rwlSlim.ExitWriteLock();
                }
            }

            public void AddTask(Task newTask)
            {
                try
                {
                    _rwlSlim.EnterWriteLock();
                    // 既にタスクが割り当てられていた場合(名前によって識別します)、
                    // そのうち誰か他の人が同じタスクを割り当てられることを期待しつつ
                    // とりあえずタスクを受け取ります。
                    var taskQuery = from t in _tasks
                                    where t == newTask
                                    select t;
                    if (taskQuery.Count<Task>() == 0)
                    {
                        Console.WriteLine("タスク " + newTask.Name + " が追加されました。");
                        _tasks.Add(newTask);
                    }
                }
                finally
                {
                    _rwlSlim.ExitWriteLock();
                }
            }

            /// <summary>
            /// Increase the priority of the task
            /// </summary>
            /// <param name="taskName">name of the task</param>
            public void IncreasePriority(string taskName)
            {
                try
                {
                    _rwlSlim.EnterUpgradeableReadLock();
                    var taskQuery = from t in _tasks
                                    where t.Name == taskName
                                    select t;
                    if (taskQuery.Count<Task>() > 0)
                    {
                        Task task = taskQuery.First<Task>();
                        _rwlSlim.EnterWriteLock();
                        task.Priority++;
                        Console.WriteLine("タスク " + task.Name +
                            " の優先度は " + task.Priority +
                            " に変更されました。");
                        _rwlSlim.ExitWriteLock();
                    }
                }
                finally
                {
                    _rwlSlim.ExitUpgradeableReadLock();
                }
            }

            /// <summary>
            /// タスクが完了したかどうかを確認できます
            /// </summary>
            /// <param name="taskName">タスクの名前</param>
            /// <returns>タスクが未完了、または登録されていなければfalse、そうでなければtrue。</returns>
            public bool IsTaskDone(string taskName)
            {
                try
                {
                    _rwlSlim.EnterReadLock();
                    var taskQuery = from t in _tasks
                                    where t.Name == taskName
                                    select t;
                    if (taskQuery.Count<Task>() > 0)
                    {
                        Task task = taskQuery.First<Task>();
                        Console.WriteLine("タスク " + task.Name + " の進捗が報告されました。");
                        return task.Status;
                    }
                }
                finally
                {
                    _rwlSlim.ExitReadLock();
                }
                return false;
            }

            private void ExecuteTask()
            {
                // 最も優先度が高いタスクを探します
                var queryResult = from t in _tasks
                                  where t.Status == false
                                  orderby t.Priority
                                  select t;
                if (queryResult.Count<Task>() > 0)
                {
                    // タスクを消化します
                    Task task = queryResult.First<Task>();
                    task.Status = true;
                    task.Priority = -1;
                    Console.WriteLine("タスク " + task.Name + " が完了しました。");
                }
            }
        }

        public class TaskManager : TaskDependent
        {
            private System.Threading.Timer _mgrTimer;

            public TaskManager(string name, Developer taskExecutor) :
                base(name, taskExecutor)
            {
                // 時間間隔を2秒に設定します
                _mgrTimer = new Timer(new TimerCallback(Intervene), null, 0, 2000);
            }

            ~TaskManager()
            {
                _mgrTimer.Dispose();
            }

            // 計画に割り込ませます
            protected void Intervene(Object stateInfo)
            {
                ChangePriority();
                // Developerが不在の場合はタイマを停止します
                if (_end)
                {
                    _mgrTimer.Dispose();
                    _developer = null;
                }
            }

            public void ChangePriority()
            {
                if (_tasks.Count > 0)
                {
                    int taskIndex = _rnd.Next(0, _tasks.Count - 1);
                    Task checkTask = _tasks[taskIndex];
                    // 無作為に選出したタスクの優先度を上げます
                    if (_developer != null)
                    {
                        _developer.IncreasePriority(checkTask.Name);
                        Console.WriteLine("{0}の介入によって、タスク {1} の優先度が変更されました", Name, checkTask.Name);
                    }
                }
            }
        }

        public class TaskDependent
        {
            protected List<Task> _tasks = new List<Task>();
            protected Developer _developer;
            protected Random _rnd = new Random();
            private Timer _taskTimer;
            private Timer _statusTimer;

            public TaskDependent(string name, Developer taskExecutor)
            {
                Name = name;
                _developer = taskExecutor;
                // 1秒おきにタスクを追加します
                _taskTimer = new Timer(new TimerCallback(AddWork), null, 0, 1000);
                // 3秒おきに進捗を確認します
                _statusTimer = new Timer(new TimerCallback(CheckStatus), null, 0, 3000);
            }

            ~TaskDependent()
            {
                _taskTimer.Dispose();
                _statusTimer.Dispose();
            }

            // タスクをDeveloperに割り当てます
            protected void AddWork(Object stateInfo)
            {
                SubmitTask();
                // Developerが不在の場合はタイマを停止します
                if (_end)
                {
                    _taskTimer.Dispose();
                    _developer = null;
                }
            }

            // Developerの進捗を確認します
            protected void CheckStatus(Object stateInfo)
            {
                CheckTaskStatus();
                // Developerが不在の場合はタイマを停止します
                if (_end)
                {
                    _statusTimer.Dispose();
                    _developer = null;
                }
            }

            public string Name { get; set; }

            public void SubmitTask()
            {
                int taskId = _rnd.Next(10000);
                string taskName = string.Format("({0}のタスク{1})", Name, taskId);
                Task newTask = new Task(taskName);
                if (_developer != null)
                {
                    _developer.AddTask(newTask);
                    _tasks.Add(newTask);
                }
            }

            public void CheckTaskStatus()
            {
                if (_tasks.Count > 0)
                {
                    int taskIndex = _rnd.Next(0, _tasks.Count - 1);
                    Task checkTask = _tasks[taskIndex];
                    if (_developer != null &&
                        _developer.IsTaskDone(checkTask.Name))
                    {
                        Console.WriteLine("{0} のタスク {1} は完了しました。", Name, checkTask.Name);
                        // TODOリストから削除します
                        _tasks.Remove(checkTask);
                    }
                }
            }
        }

        #endregion // "11.11 - 読み取り主体のアクセスを効率的に実行する"
    }
}
