﻿using System;
using System.ComponentModel;
using System.IO;
using System.Runtime.InteropServices;
using System.Runtime.Serialization.Formatters.Binary;
using System.Threading;

namespace MutexFun
{
    /// <summary>
    /// Mutexオブジェクトを使用して共有メモリアクセスの同期をとることにより
    /// 共有メモリ経由でデータを送信するクラスです。
    /// </summary>
    public class SharedMemoryManager<TransferItemType> : IDisposable
    {
        #region Consts
        const int INVALID_HANDLE_VALUE = -1;
        const int FILE_MAP_WRITE = 0x0002;
        /// <summary>
        /// Win32 APIの定義に従っています。
        /// </summary>
        const int ERROR_ALREADY_EXISTS = 183;
        #endregion

        #region Private members
        IntPtr _handleFileMapping = IntPtr.Zero;
        IntPtr _ptrToMemory = IntPtr.Zero;
        uint _memRegionSize = 0;
        string _memoryRegionName;
        bool disposed = false;
        int _sharedMemoryBaseSize = 0;
        Mutex _mtxSharedMem = null;

        #endregion

        #region コンストラクタ / 後処理
        public SharedMemoryManager(string name, int sharedMemoryBaseSize)
        {
            // シリアル化可能なオブジェクトだけをサポートします
            if (!typeof(TransferItemType).IsSerializable)
                throw new ArgumentException(
                    string.Format("オブジェクト {0} はシリアル化可能ではありません。",
                        typeof(TransferItemType)));

            if (string.IsNullOrEmpty(name))
                throw new ArgumentNullException("name");

            if (sharedMemoryBaseSize <= 0)
                throw new ArgumentOutOfRangeException("sharedMemoryBaseSize",
                    "共有メモリの基本サイズは0以上である必要があります。");

            // 領域の名前を設定します
            _memoryRegionName = name;
            // 基本サイズを保存します
            _sharedMemoryBaseSize = sharedMemoryBaseSize;
            // メモリ領域のサイズを設定します
            _memRegionSize = (uint)(_sharedMemoryBaseSize + sizeof(int));
            // 共有メモリセクションを設定します
            SetupSharedMemory();
        }

        private void SetupSharedMemory()
        {
            // ページファイルからストレージ領域を確保します
            _handleFileMapping =
                PInvoke.CreateFileMapping((IntPtr)INVALID_HANDLE_VALUE,
                                    IntPtr.Zero,
                                    PInvoke.PageProtection.ReadWrite,
                                    0,
                                    _memRegionSize,
                                    _memoryRegionName);

            if (_handleFileMapping == IntPtr.Zero)
            {
                throw new Win32Exception(
                    "ファイルマッピングが作成できませんでした");
            }

            // エラーを確認します
            int retVal = Marshal.GetLastWin32Error();
            if (retVal == ERROR_ALREADY_EXISTS)
            {
                // 別の所有者が存在する既存のMutexオブジェクトを
                // 取得しようとしました。そのため
                // 既存のMutexオブジェクトへ接続します。
                _mtxSharedMem = new Mutex(false,
                    string.Format("{0}mtx{1}",
                        typeof(TransferItemType), _memoryRegionName));
            }
            else if (retVal == 0)
            {
                // 新規Mutexオブジェクトをオープンしました。
                // 最初の所有者としてMutexオブジェクトを作成します。
                _mtxSharedMem = new Mutex(true,
                    string.Format("{0}mtx{1}",
                        typeof(TransferItemType), _memoryRegionName));
            }
            else
            {
                // 予期しないエラーが発生しました
                throw new Win32Exception(retVal, "ファイルマッピングの作成中にエラーが発生しました。");
            }

            // 共有メモリにマップします
            _ptrToMemory = PInvoke.MapViewOfFile(_handleFileMapping,
                                            FILE_MAP_WRITE,
                                            0, 0, IntPtr.Zero);

            if (_ptrToMemory == IntPtr.Zero)
            {
                retVal = Marshal.GetLastWin32Error();
                throw new Win32Exception(retVal, "マップファイルを参照できませんでした。");
            }

            retVal = Marshal.GetLastWin32Error();
            if (retVal != 0 && retVal != ERROR_ALREADY_EXISTS)
            {
                // 予期しないエラーが発生しました。
                throw new Win32Exception(retVal, "ファイルのマッピング中にエラーが発生しました。");
            }
        }

        ~SharedMemoryManager()
        {
            // メモリのクローズを確認します
            Dispose(false);
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        private void Dispose(bool disposing)
        {
            // Disposeが既に呼ばれたかどうかを確認します
            if (!this.disposed)
            {
                CloseSharedMemory();
            }
            disposed = true;
        }

        private void CloseSharedMemory()
        {
            if (_ptrToMemory != IntPtr.Zero)
            {
                // 共有メモリ上のマッピングをクローズします
                PInvoke.UnmapViewOfFile(_ptrToMemory);
                _ptrToMemory = IntPtr.Zero;
            }
            if (_handleFileMapping != IntPtr.Zero)
            {
                // ハンドルをクローズします
                PInvoke.CloseHandle(_handleFileMapping);
                _handleFileMapping = IntPtr.Zero;
            }
        }

        public void Close()
        {
            CloseSharedMemory();
        }
        #endregion

        #region プロパティ
        public int SharedMemoryBaseSize
        {
            get { return _sharedMemoryBaseSize; }
        }
        #endregion

        #region Public メソッド
        /// <summary>
        /// 共有メモリを経由することにより、シリアル化可能なオブジェクトを送信します。
        /// また、オブジェクトが受け取られるまで待機します。
        /// </summary>
        /// <param name="transferObject"></param>
        public void SendObject(TransferItemType transferObject)
        {
            // MemoryStreamを作成してサイズを初期化します
            using (MemoryStream ms = new MemoryStream())
            {
                // シリアル化用のフォーマッタを用意します
                BinaryFormatter formatter = new BinaryFormatter();
                try
                {
                    // オブジェクトをストリーム上にシリアル化します
                    formatter.Serialize(ms, transferObject);

                    // シリアル化したオブジェクトのサイズを取得します
                    byte[] bytes = ms.ToArray();

                    // オブジェクトのサイズが適当かどうか確認します
                    if (bytes.Length + sizeof(int) > _memRegionSize)
                    {
                        string fmt =
                            "{0} オブジェクトのインスタンスをシリアル化した後のサイズ {1} バイトは" +
                            "共有メモリ領域のサイズよりも大きいサイズです。";

                        string msg =
                            string.Format(fmt,
                                typeof(TransferItemType), bytes.Length);

                        throw new ArgumentException(msg, "transferObject");
                    }

                    // オブジェクトのサイズを書き込みます
                    Marshal.WriteInt32(this._ptrToMemory, bytes.Length);

                    // データを表すバイト列を書き込みます
                    Marshal.Copy(bytes, 0, this._ptrToMemory, bytes.Length);
                }
                finally
                {
                    // Mutexを使用する他のプロセスにシグナルを送信して
                    // 受信処理を行えるようにします
                    _mtxSharedMem.ReleaseMutex();

                    // 他のプロセスがデータを受信したことを確認するために
                    // 待機した後、処理を継続します
                    _mtxSharedMem.WaitOne();
                }
            }
        }

        /// <summary>
        /// 共有メモリ上にオブジェクトが保存されたかどうか確認して、保存されていれば逆シリアル化を行って
        /// データを受け取ります。
        /// </summary>
        /// <returns>送信されたオブジェクト</returns>
        public TransferItemType ReceiveObject()
        {
            // 送信者によってデータが保存されるまで待機します
            _mtxSharedMem.WaitOne();

            // 共有メモリのサイズを取得します
            int count = Marshal.ReadInt32(_ptrToMemory);
            if (count <= 0)
            {
                throw new InvalidDataException("読み取るデータがありません");
            }

            // データを保持するためのbyte配列を用意します
            byte[] bytes = new byte[count];

            // 送信されたオブジェクトをbyte配列として取得します
            Marshal.Copy(_ptrToMemory, bytes, 0, count);

            // byte配列を読み取るためのMemoryStreamを作成します
            using (MemoryStream ms = new MemoryStream(bytes))
            {
                // バイナリフォーマッタを用意します
                BinaryFormatter formatter = new BinaryFormatter();

                // 返り値となるオブジェクトを作成します
                TransferItemType item;
                try
                {
                    item = (TransferItemType)formatter.Deserialize(ms);
                }
                finally
                {
                    // Mutexを使用してデータを受け取ったことを通知します
                    _mtxSharedMem.ReleaseMutex();
                }
                // 取得したオブジェクトを返します
                return item;
            }
        }
        #endregion
    }

    public class PInvoke
    {
        #region PInvoke defines
        [Flags]
        public enum PageProtection : uint
        {
            NoAccess = 0x01,
            Readonly = 0x02,
            ReadWrite = 0x04,
            WriteCopy = 0x08,
            Execute = 0x10,
            ExecuteRead = 0x20,
            ExecuteReadWrite = 0x40,
            ExecuteWriteCopy = 0x80,
            Guard = 0x100,
            NoCache = 0x200,
            WriteCombine = 0x400,
        }

        [DllImport("kernel32.dll", SetLastError = true)]
        public static extern IntPtr CreateFileMapping(IntPtr hFile,
            IntPtr lpFileMappingAttributes, PageProtection flProtect,
            uint dwMaximumSizeHigh,
            uint dwMaximumSizeLow, string lpName);

        [DllImport("kernel32.dll", SetLastError = true)]
        public static extern IntPtr MapViewOfFile(IntPtr hFileMappingObject, uint
            dwDesiredAccess, uint dwFileOffsetHigh, uint dwFileOffsetLow,
            IntPtr dwNumberOfBytesToMap);

        [DllImport("kernel32.dll", SetLastError = true)]
        public static extern bool UnmapViewOfFile(IntPtr lpBaseAddress);

        [DllImport("kernel32.dll", SetLastError = true)]
        public static extern bool CloseHandle(IntPtr hObject);
        #endregion
    }
}
